from ..events import (
    event_document_trashed, event_trashed_document_deleted,
    event_trashed_document_restored
)
from ..models.document_models import Document
from ..models.trashed_document_models import TrashedDocument
from ..permissions import (
    permission_document_trash, permission_document_view,
    permission_trash_empty, permission_trashed_document_delete,
    permission_trashed_document_restore
)

from .base import GenericDocumentViewTestCase
from .mixins.trashed_document_mixins import TrashedDocumentViewTestMixin


class DocumentTrashViewTestCase(
    TrashedDocumentViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_trash_get_view_no_permission(self):
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_document_trash_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_trash_get_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_trash
        )

        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_document_trash_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_trash_get_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_trash
        )

        self._test_document.delete()

        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_document_trash_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_trash_post_view_no_permission(self):
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_document_trash_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_trash_post_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_trash
        )

        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_document_trash_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.valid.count(), document_count - 1)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_trashed.id)

    def test_trashed_document_trash_post_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_trash
        )

        self._test_document.delete()

        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_document_trash_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class TrashedDocumentViewTestCase(
    TrashedDocumentViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_trashed_document_delete_get_view_no_permission(self):
        self._test_document.delete()

        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_trashed_document_delete_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.valid.count(), document_count
        )
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_delete_get_view_with_access(self):
        self._test_document.delete()
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_delete
        )

        self._clear_events()

        response = self._request_test_trashed_document_delete_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            Document.valid.count(), document_count
        )
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_delete_post_view_no_permission(self):
        self._test_document.delete()
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_trashed_document_delete_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_delete_post_view_with_access(self):
        self._test_document.delete()
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_delete
        )

        self._clear_events()

        response = self._request_test_trashed_document_delete_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(events[0].verb, event_trashed_document_deleted.id)

    def test_trashed_document_list_view_no_permission(self):
        self._test_document.delete()

        self._clear_events()

        response = self._request_test_trashed_document_list_view()
        self.assertNotContains(
            response=response, text=self._test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_list_view_with_access(self):
        self._test_document.delete()

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_trashed_document_list_view()
        self.assertContains(
            response=response, text=self._test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_get_view_no_permission(self):
        self._test_document.delete()
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_get_view_with_access(self):
        self._test_document.delete()
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_restore
        )

        document_count = Document.valid.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_post_view_no_permission(self):
        self._test_document.delete()
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_test_trashed_document_restore_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_restore_post_view_with_access(self):
        self._test_document.delete()
        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_restore
        )

        self._clear_events()

        response = self._request_test_trashed_document_restore_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.valid.count(), document_count + 1)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_trashed_document_restored.id)


class TrashCanViewTestCase(
    TrashedDocumentViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_trash_can_empty_view_no_permission(self):
        self._test_document.delete()

        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self._clear_events()

        response = self._request_empty_trash_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trash_can_empty_view_with_permission(self):
        self._test_document.delete()

        document_count = Document.valid.count()
        trashed_document_count = TrashedDocument.objects.count()

        self.grant_permission(permission=permission_trash_empty)

        self._clear_events()

        response = self._request_empty_trash_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.valid.count(), document_count)
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(events[0].verb, event_trashed_document_deleted.id)



from ..document_file_actions import (
    DocumentFileActionAppendNewPages, DocumentFileActionNothing,
    DocumentFileActionUseNewPages
)
from ..events import (
    event_document_file_created, event_document_file_edited,
    event_document_version_created, event_document_version_edited,
    event_document_version_page_created
)

from .base import GenericDocumentTestCase
from .mixins.document_file_mixins import DocumentFileTestMixin
from .mixins.document_version_mixins import DocumentVersionTestMixin


class DocumentFileActionTestCase(
    DocumentFileTestMixin, GenericDocumentTestCase
):
    def test_version_new_file_new_pages(self):
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()

        test_document_version_page_content_objects = self._test_document_version.page_content_objects

        self.assertEqual(
            self._test_document.versions.count(), 1
        )

        self._clear_events()

        self._upload_test_document_file(
            action_name=DocumentFileActionUseNewPages.backend_id,
            user=self._test_case_user
        )

        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count + 1
        )

        self.assertEqual(
            self._test_document_version_list[0].active, False
        )
        self.assertEqual(
            self._test_document_version_list[1].active, True
        )

        self.assertNotEqual(
            self._test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertEqual(
            self._test_document_version.page_content_objects,
            list(
                self._test_document.file_latest.pages.all()
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(
            events[3].action_object, self._test_document_version
        )
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[4].action_object, self._test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version)
        self.assertEqual(events[4].verb, event_document_version_edited.id)

    def test_version_new_version_keep_pages(self):
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()

        test_document_version_page_content_objects = self._test_document_version.page_content_objects

        self.assertEqual(
            self._test_document.versions.count(), 1
        )

        self._clear_events()

        self._upload_test_document_file(
            action_name=DocumentFileActionNothing.backend_id,
            user=self._test_case_user
        )

        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count
        )

        self.assertEqual(
            self._test_document_version_list[0].active, True
        )

        self.assertEqual(
            self._test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertNotEqual(
            self._test_document_version.page_content_objects,
            list(
                self._test_document.file_latest.pages.all()
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

    def test_version_new_file_append_pages(self):
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()

        test_document_version_page_content_objects = self._test_document_version.page_content_objects

        self._clear_events()

        self._upload_test_document_file(
            action_name=DocumentFileActionAppendNewPages.backend_id,
            user=self._test_case_user
        )

        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count + 1
        )

        test_document_version_expected_page_content_objects = list(
            self._test_document.files.first().pages.all()
        )
        test_document_version_expected_page_content_objects.extend(
            list(
                self._test_document.files.last().pages.all()
            )
        )

        self.assertEqual(self._test_document_version_list[0].active, False)
        self.assertEqual(self._test_document_version_list[1].active, True)

        self.assertEqual(
            self._test_document_version_list[0].page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertEqual(
            self._test_document_version_list[1].page_content_objects,
            test_document_version_expected_page_content_objects
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(
            events[3].action_object, self._test_document_version
        )
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, self._test_document_version_page_list[1]
        )
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, self._test_document_version_page_list[2]
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)


class DocumentVersionTestCase(
    DocumentVersionTestMixin, GenericDocumentTestCase
):
    def test_create(self):
        self._clear_events()

        self._create_test_document_version(user=self._test_case_user)

        self.assertEqual(self._test_document_version_list[0].active, True)
        self.assertEqual(self._test_document_version_list[1].active, False)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].target, self._test_document_version_list[1]
        )
        self.assertEqual(events[0].verb, event_document_version_created.id)

    def test_method_get_absolute_url(self):
        self._clear_events()

        self.assertTrue(
            self._test_document.version_active.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentVersionBusinessLogicTestCase(
    DocumentVersionTestMixin, GenericDocumentTestCase
):
    def test_multiple_active(self):
        self._create_test_document_version(user=self._test_case_user)

        self._test_document_version_list[0].refresh_from_db()
        self._test_document_version_list[1].refresh_from_db()

        self.assertEqual(
            self._test_document_version_list[0].active, True
        )
        self.assertEqual(
            self._test_document_version_list[1].active, False
        )

        self._clear_events()

        self._test_document_version_list[1].active = True
        self._test_document_version_list[1]._event_actor = self._test_case_user
        self._test_document_version_list[1].save()

        self._test_document_version_list[0].refresh_from_db()
        self._test_document_version_list[1].refresh_from_db()

        self.assertEqual(
            self._test_document_version_list[0].active, False
        )
        self.assertEqual(
            self._test_document_version_list[1].active, True
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].target, self._test_document_version_list[1]
        )
        self.assertEqual(events[0].verb, event_document_version_edited.id)



import itertools

from django.apps import apps
from django.contrib.admin.utils import reverse_field_path
from django.db.models.aggregates import Max, Min
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.utils import group_iterator, parse_range
from mayan.apps.databases.literals import DATABASE_MINIMUM_ID
from mayan.apps.views.literals import LIST_MODE_CHOICE_LIST

from .exceptions import DynamicSearchException
from .literals import QUERY_PARAMETER_ANY_FIELD
from .search_fields import SearchField
from .settings import setting_indexing_chunk_size


class SearchModel(AppsModuleLoaderMixin):
    _loader_module_name = 'search'
    _registry = {}

    @staticmethod
    def function_return_same(value):
        return value

    @classmethod
    def all(cls):
        result = set(
            cls._registry.values()
        )
        result = list(result)
        result.sort(key=lambda entry: entry.label)
        return result

    @classmethod
    def get(cls, name):
        try:
            result = cls._registry[name]
        except KeyError:
            raise KeyError(
                _(message='Unknown search model `%s`.') % name
            )
        else:
            if getattr(result, 'serializer_path', None):
                result.serializer = import_string(
                    dotted_path=result.serializer_path
                )

        return result

    @classmethod
    def get_default(cls):
        for search_class in cls.all():
            if search_class.default:
                return search_class

    @classmethod
    def get_for_model(cls, instance):
        # Works the same for model classes and model instances.
        return cls.get(
            name=instance._meta.label.lower()
        )

    @classmethod
    def get_through_models(cls):
        through_models = {}

        for search_model in cls.all():
            for related_model, reverse_field_path_text in search_model.get_related_models():
                # Check is each related model is connected to a many to many.
                for field in related_model._meta.get_fields():
                    if field.many_to_many:
                        try:
                            through_model = field.through
                        except AttributeError:
                            through_model = field.remote_field.through

                        through_models.setdefault(
                            through_model, {}
                        )
                        through_models[through_model].setdefault(
                            related_model, set()
                        )
                        through_models[through_model][related_model].add(
                            reverse_field_path_text
                        )

        return through_models

    def __init__(
        self, app_label, model_name, default=False, label=None,
        list_mode=None, manager_name=None, permission=None,
        queryset=None, serializer_path=None
    ):
        self.default = default
        self._label = label
        self.app_label = app_label
        self.list_mode = list_mode or LIST_MODE_CHOICE_LIST
        self.model_name = model_name.lower()
        self._proxies = []  # Lazy evaluation.
        self.permission = permission
        self.queryset = queryset
        self.search_fields_dict = {}
        self.serializer_path = serializer_path

        auto_field = self.base_model._meta.auto_field
        self.add_model_field(
            field=auto_field.name, label=auto_field.verbose_name
        )
        self.add_model_field(
            field=QUERY_PARAMETER_ANY_FIELD, label=_(message='All content')
        )

        self.manager_name = manager_name or self.model._meta.default_manager.name

        if default:
            for search_class in self.__class__._registry.values():
                search_class.default = False

        self.__class__._registry[self.full_name] = self

    def __repr__(self):
        return '<{}: {}>'.format(
            self.__class__.__name__, self.label
        )

    def __str__(self):
        return str(self.label)

    def add_model_field(self, **kwargs):
        """
        Add a search field that directly belongs to the parent SearchModel.
        """
        self.do_search_field_cache_invalidate()

        kwargs['search_model'] = self
        search_field = SearchField.init(**kwargs)
        self.search_fields_dict[search_field.field_name] = search_field
        return search_field

    def add_proxy_model(self, app_label, model_name):
        model_name = model_name.lower()
        self._proxies.append(
            {
                'app_label': app_label, 'model_name': model_name
            }
        )

        self.__class__._registry[
            '{}.{}'.format(app_label, model_name)
        ] = self

    @cached_property
    def base_model(self):
        return self.model._meta.proxy_for_model or self.model

    def do_search_field_cache_invalidate(self):
        self.__dict__.pop('search_field_name_list', None)
        self.__dict__.pop('search_fields', None)

    @cached_property
    def full_name(self):
        return '{}.{}'.format(self.app_label, self.model_name)

    def get_id_groups(self, range_string=None):
        """
        Generate ID groups when doing bulk indexing. ID groups avoid having
        to create a single task call for each object to be indexed.
        """
        queryset = self.model._meta.managers_map[self.manager_name].all()

        # Part 1 - Split the user requested range into blind groups.
        if not range_string:
            # If range is not specified it will be the minimum and maximum
            # IDs of the queryset.
            queryset_id_values = queryset.aggregate(
                min_id=Min('id'), max_id=Max('id')
            )
            range_string = '{}-{}'.format(
                queryset_id_values['min_id'] or DATABASE_MINIMUM_ID,
                queryset_id_values['max_id'] or DATABASE_MINIMUM_ID
            )

        # Part 2 - Validate the blind groups by querying them and retrieve
        # the valid ID values.
        id_list_groups = group_iterator(
            iterable=parse_range(range_string=range_string),
            group_size=setting_indexing_chunk_size.value
        )

        generator_valid_id_groups = (
            queryset.filter(
                pk__in=id_list
            ).values_list('id', flat=True) for id_list in id_list_groups
        )

        # Part 3 - Chain the valid ID groups into a single sequence and
        # split them again into groups.
        return group_iterator(
            iterable=itertools.chain.from_iterable(
                generator_valid_id_groups
            ), group_size=setting_indexing_chunk_size.value
        )

    def get_queryset(self):
        if self.queryset is not None:
            return self.queryset()
        else:
            return self.model._meta.managers_map[self.manager_name].all()

    def get_related_models(self):
        result = set()
        for search_field in self.search_fields:
            if search_field.concrete:
                obj, path = reverse_field_path(
                    model=self.model, path=search_field.field_name
                )
                if path:
                    # Ignore search model fields.
                    result.add(
                        (obj, path)
                    )

        return result

    def get_search_field(self, field_name):
        try:
            return self.search_fields_dict[field_name]
        except KeyError:
            raise DynamicSearchException(
                'No search field named: %s' % field_name
            )

    def get_search_field_choices(self):
        """
        Returns a list of the fields for the SearchModel.
        """
        result = []
        for search_field in self.search_fields:
            result.append(
                (search_field.field_name, search_field.label)
            )

        return sorted(
            result, key=lambda x: x[1]
        )

    @cached_property
    def label(self):
        if not self._label:
            self._label = self.model._meta.verbose_name
        return self._label

    @cached_property
    def model(self):
        return apps.get_model(
            app_label=self.app_label, model_name=self.model_name
        )

    @cached_property
    def pk(self):
        return self.full_name

    def populate(
        self, instance, search_backend, exclude_kwargs=None,
        exclude_model=None
    ):
        instance_field_data = {}

        # Process the search fields by order of priority. This makes sure
        # that virtual fields are processed last.
        for search_field in self.search_fields_priority_sorted:
            field_value = search_field.get_instance_value(
                exclude_kwargs=exclude_kwargs, exclude_model=exclude_model,
                instance=instance, instance_field_data=instance_field_data,
                search_backend=search_backend
            )
            if field_value is not None:
                instance_field_data[search_field.field_name] = field_value

        return instance_field_data

    @property
    def proxies(self):
        result = []
        for proxy in self._proxies:
            result.append(
                apps.get_model(
                    app_label=proxy['app_label'],
                    model_name=proxy['model_name']
                )
            )
        return result

    def remove_search_field(self, search_field):
        self.do_search_field_cache_invalidate()
        self.search_fields_dict.pop(search_field.field_name)

    @cached_property
    def search_field_name_list(self):
        return [
            search_field.field_name for search_field in self.search_fields
        ]

    @cached_property
    def search_fields(self):
        return self.search_fields_dict.values()

    @cached_property
    def search_fields_label_sorted(self):
        return sorted(
            self.search_fields,
            key=lambda search_field: search_field.get_label()
        )

    @cached_property
    def search_fields_priority_sorted(self):
        return sorted(
            self.search_fields,
            key=lambda search_field: search_field.priority
        )



import uuid

from django.db import connection, migrations, models


def code_change_uuid_field_type(apps, schema_editor):
    if not schema_editor.connection.vendor == 'oracle':
        # Skip this migration for Oracle
        # GitHub issue #251
        migrations.AlterField(
            model_name='document',
            name='uuid',
            field=models.UUIDField(default=uuid.uuid4, editable=False)
        )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0031_convert_uuid')
    ]

    operations = [
        migrations.RunPython(code=code_change_uuid_field_type)
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if connection.vendor == 'postgresql':
            self.operations.insert(
                0, migrations.RunSQL(
                    'ALTER TABLE documents_document ALTER COLUMN uuid SET DATA TYPE UUID USING uuid::uuid;'
                )
            )



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_signatures', '0011_rename_signaturebasemodel_field')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='detachedsignature',
            options={
                'verbose_name': 'Document file detached signature',
                'verbose_name_plural': 'Document file detached signatures'
            }
        ),
        migrations.AlterModelOptions(
            name='embeddedsignature',
            options={
                'verbose_name': 'Document file embedded signature',
                'verbose_name_plural': 'Document file embedded signatures'
            }
        ),
        migrations.AlterModelOptions(
            name='signaturebasemodel',
            options={
                'ordering': ('pk',),
                'verbose_name': 'Document file signature',
                'verbose_name_plural': 'Document file signatures'
            }
        ),
        migrations.AlterField(
            model_name='signaturebasemodel',
            name='document_file',
            field=models.ForeignKey(
                editable=False, on_delete=django.db.models.deletion.CASCADE,
                related_name='signatures', to='documents.DocumentFile',
                verbose_name='Document file'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_document_metadata_add, icon_document_metadata_edit,
    icon_document_metadata_list, icon_document_metadata_remove,
    icon_document_type_metadata_type_list, icon_metadata_type_create,
    icon_metadata_type_document_type_list, icon_metadata_type_edit,
    icon_metadata_type_list, icon_metadata_type_multiple_delete,
    icon_metadata_type_single_delete
)
from .permissions import (
    permission_document_metadata_add, permission_document_metadata_edit,
    permission_document_metadata_remove, permission_document_metadata_view,
    permission_metadata_type_create, permission_metadata_type_delete,
    permission_metadata_type_edit, permission_metadata_type_view
)

# Document metadata

link_metadata_add = Link(
    args='object.pk', icon=icon_document_metadata_add,
    permission=permission_document_metadata_add, text=_(message='Add metadata'),
    view='metadata:metadata_add',
)
link_metadata_edit = Link(
    args='object.pk', icon=icon_document_metadata_edit,
    permission=permission_document_metadata_edit,
    text=_(message='Edit metadata'), view='metadata:metadata_edit'
)
link_metadata_multiple_add = Link(
    icon=icon_document_metadata_add, text=_(message='Add metadata'),
    view='metadata:metadata_multiple_add'
)
link_metadata_multiple_edit = Link(
    icon=icon_document_metadata_edit, text=_(message='Edit metadata'),
    view='metadata:metadata_multiple_edit'
)
link_metadata_multiple_remove = Link(
    icon=icon_document_metadata_remove, text=_(message='Remove metadata'),
    view='metadata:metadata_multiple_remove'
)
link_metadata_remove = Link(
    args='object.pk', icon=icon_document_metadata_remove,
    permission=permission_document_metadata_remove,
    text=_(message='Remove metadata'), view='metadata:metadata_remove',
)
link_metadata_list = Link(
    args='resolved_object.pk', icon=icon_document_metadata_list,
    permission=permission_document_metadata_view, text=_(message='Metadata'),
    view='metadata:metadata_list',
)

# Document type

link_document_type_metadata_type_relationship = Link(
    args='resolved_object.pk',
    icon=icon_document_type_metadata_type_list,
    permission=permission_document_type_edit,
    text=_(message='Metadata types'), view='metadata:document_type_metadata_type_relationship',
)

# Metadata type

link_metadata_type_document_type_relationship = Link(
    args='resolved_object.pk',
    icon=icon_metadata_type_document_type_list,
    permission=permission_document_type_edit,
    text=_(message='Document types'), view='metadata:metadata_type_document_type_relationship',
)
link_metadata_type_create = Link(
    icon=icon_metadata_type_create,
    permission=permission_metadata_type_create, text=_(message='Create new'),
    view='metadata:metadata_type_create'
)
link_metadata_type_single_delete = Link(
    args='object.pk', icon=icon_metadata_type_single_delete,
    permission=permission_metadata_type_delete,
    tags='dangerous', text=_(message='Delete'),
    view='metadata:metadata_type_single_delete',
)
link_metadata_type_multiple_delete = Link(
    icon=icon_metadata_type_multiple_delete,
    text=_(message='Delete'), view='metadata:metadata_type_multiple_delete'
)
link_metadata_type_edit = Link(
    args='object.pk', icon=icon_metadata_type_edit,
    permission=permission_metadata_type_edit,
    text=_(message='Edit'), view='metadata:metadata_type_edit'
)
link_metadata_type_list = Link(
    icon=icon_metadata_type_list, text=_(message='Metadata types'),
    view='metadata:metadata_type_list'
)
link_metadata_type_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='metadata', model_name='MetadataType',
        object_permission=permission_metadata_type_view,
        view_permission=permission_metadata_type_create,
    ), icon=icon_metadata_type_list, text=_(message='Metadata types'),
    view='metadata:metadata_type_list'
)



def load_env_file(filename='config.env', skip_local_config=False):
    result = {}
    with open(file=filename) as file_object:
        for line in file_object:
            if not line.startswith('#'):
                key, value = line.strip().split('=')

                result[key] = value

    if filename != 'config-local.env' and not skip_local_config:
        try:
            result.update(
                load_env_file(filename='config-local.env')
            )
        except FileNotFoundError:
            """
            Non fatal. Just means this deployment does not overrides the
            default `config.env` file values.
            """

    return result



from django.dispatch import Signal

signal_post_document_type_change = Signal(use_caching=True)
signal_post_initial_document_type = Signal(use_caching=True)
signal_post_document_file_upload = Signal(use_caching=True)
signal_post_document_version_remap = Signal(use_caching=True)



from django import forms
from django.forms.formsets import formset_factory
from django.utils.translation import gettext_lazy as _

from mayan.apps.templating.fields import ModelTemplateField

from ..models import WorkflowInstance, WorkflowTransition


class WorkflowTransitionForm(forms.ModelForm):
    def __init__(self, workflow, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields[
            'origin_state'
        ].queryset = self.fields[
            'origin_state'
        ].queryset.filter(workflow=workflow)

        self.fields[
            'destination_state'
        ].queryset = self.fields[
            'destination_state'
        ].queryset.filter(workflow=workflow)

        self.fields['condition'] = ModelTemplateField(
            initial_help_text=self.fields['condition'].help_text,
            label=self.fields['condition'].label, model=WorkflowInstance,
            model_variable='workflow_instance', required=False
        )

    class Meta:
        fields = ('label', 'origin_state', 'destination_state', 'condition')
        model = WorkflowTransition


class WorkflowTransitionTriggerEventRelationshipForm(forms.Form):
    namespace = forms.CharField(
        label=_(message='Namespace'), required=False,
        widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    label = forms.CharField(
        label=_(message='Label'), required=False,
        widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    relationship = forms.ChoiceField(
        choices=(
            ('no', _(message='No')),
            ('yes', _(message='Yes')),
        ), label=_(message='Enabled'), widget=forms.RadioSelect()
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['namespace'].initial = self.initial['event_type'].namespace
        self.fields['label'].initial = self.initial['event_type'].label

        relationship = self.initial['transition'].trigger_events.filter(
            event_type=self.initial['event_type']
        )

        if relationship.exists():
            self.fields['relationship'].initial = 'yes'
        else:
            self.fields['relationship'].initial = 'no'

    def save(self):
        relationship = self.initial['transition'].trigger_events.filter(
            event_type=self.initial['event_type']
        )

        if self.cleaned_data['relationship'] == 'no':
            relationship.delete()
        elif self.cleaned_data['relationship'] == 'yes':
            if not relationship.exists():
                self.initial['transition'].trigger_events.create(
                    event_type=self.initial['event_type']
                )


WorkflowTransitionTriggerEventRelationshipFormSet = formset_factory(
    form=WorkflowTransitionTriggerEventRelationshipForm, extra=0
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Smart links'), name='linking'
)

permission_resolved_smart_link_view = namespace.add_permission(
    label=_(message='View resolved smart links'), name='resolved_smart_link_view'
)
permission_smart_link_create = namespace.add_permission(
    label=_(message='Create new smart links'), name='smart_link_create'
)
permission_smart_link_delete = namespace.add_permission(
    label=_(message='Delete smart links'), name='smart_link_delete'
)
permission_smart_link_edit = namespace.add_permission(
    label=_(message='Edit smart links'), name='smart_link_edit'
)
permission_smart_link_view = namespace.add_permission(
    label=_(message='View existing smart links'), name='smart_link_view'
)



from django.apps import apps
from django.db import models


class MetadataTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

    def get_for_document(self, document):
        return self.filter(
            pk__in=document.metadata.values_list(
                'metadata_type'
            )
        )

    def get_for_document_type(self, document_type, required=None):
        queryset_document_type_metadata_types = document_type.metadata.all()

        if required is not None:
            queryset_document_type_metadata_types = queryset_document_type_metadata_types.filter(
                required=required
            )

        return self.filter(
            pk__in=queryset_document_type_metadata_types.values_list(
                'metadata_type'
            )
        )

    def get_for_document_types(self, queryset):
        return self.filter(
            document_types__document_type__in=queryset.values('pk')
        ).distinct()


class DocumentTypeMetadataTypeManager(models.Manager):
    def get_by_natural_key(
        self, document_natural_key, metadata_type_natural_key
    ):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        MetadataType = apps.get_model(
            app_label='metadata', model_name='MetadataType'
        )
        try:
            document = Document.objects.get_by_natural_key(
                document_natural_key
            )
        except Document.DoesNotExist:
            raise self.model.DoesNotExist
        else:
            try:
                metadata_type = MetadataType.objects.get_by_natural_key(metadata_type_natural_key)
            except MetadataType.DoesNotExist:
                raise self.model.DoesNotExist

        return self.get(
            document__pk=document.pk, metadata_type__pk=metadata_type.pk
        )



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..classes import ModelPermission
from ..events import event_acl_created, event_acl_deleted, event_acl_edited
from ..models import AccessControlList
from ..permissions import permission_acl_edit, permission_acl_view

from .mixins import ACLAPIViewTestMixin, ACLTestMixin


class ACLAPIViewTestCase(ACLTestMixin, ACLAPIViewTestMixin, BaseAPITestCase):
    auto_create_acl_test_object = True

    def test_acl_create_api_view_no_permission(self):
        acl_count = AccessControlList.objects.count()

        self._clear_events()

        response = self._request_test_acl_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(AccessControlList.objects.count(), acl_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_create_api_view_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_acl_edit
        )

        self._clear_events()

        response = self._request_test_acl_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self._test_acl.role, self._test_role)
        self.assertEqual(self._test_acl.content_object, self._test_object)
        self.assertEqual(self._test_acl.permissions.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_acl)
        self.assertEqual(events[0].verb, event_acl_created.id)

    def test_acl_delete_api_view_no_permission(self):
        self._create_test_acl()

        acl_count = AccessControlList.objects.count()

        self._clear_events()

        response = self._request_test_acl_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(AccessControlList.objects.count(), acl_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_delete_api_view_with_access(self):
        self._create_test_acl()

        self.grant_access(self._test_object, permission=permission_acl_edit)

        acl_count = AccessControlList.objects.count()

        self._clear_events()

        response = self._request_test_acl_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(AccessControlList.objects.count(), acl_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_object)
        self.assertEqual(events[0].verb, event_acl_deleted.id)

    def test_acl_detail_api_view_no_permission(self):
        self._create_test_acl()

        self._clear_events()

        response = self._request_test_acl_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_detail_api_view_with_access(self):
        self._create_test_acl()

        self.grant_access(
            obj=self._test_object, permission=permission_acl_view
        )

        self._clear_events()

        response = self._request_test_acl_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['content_type']['app_label'],
            self._test_object_content_type.app_label
        )
        self.assertEqual(
            response.data['role']['label'], self._test_role.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_list_api_view_no_permission(self):
        self._create_test_acl()

        self._clear_events()

        response = self._request_test_acl_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_list_api_view_with_access(self):
        self._create_test_acl()

        self.grant_access(
            obj=self._test_object, permission=permission_acl_view
        )

        self._clear_events()

        response = self._request_test_acl_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertContains(
            response=response, text=self._test_object_content_type.app_label
        )
        self.assertContains(
            response=response, text=self._test_acl.role.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class ACLPermissionAPIViewTestCase(
    ACLTestMixin, ACLAPIViewTestMixin, BaseAPITestCase
):
    auto_create_acl_test_object = True

    def test_acl_permission_add_api_view_no_permission(self):
        self._create_test_acl()

        self._clear_events()

        response = self._request_test_acl_permission_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(
            self._test_permission.stored_permission in self._test_acl.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_permission_add_api_view_with_access(self):
        self._create_test_acl()

        self.grant_access(
            obj=self._test_object, permission=permission_acl_edit
        )

        self._clear_events()

        response = self._request_test_acl_permission_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(
            self._test_permission.stored_permission in self._test_acl.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_acl)
        self.assertEqual(events[0].verb, event_acl_edited.id)

    def test_acl_permission_list_api_view_no_permission(self):
        self._create_test_acl()
        self._test_acl.permissions.add(
            self._test_permission.stored_permission
        )

        self._clear_events()

        response = self._request_test_acl_permission_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_permission_list_api_view_with_access(self):
        self._create_test_acl()
        self._test_acl.permissions.add(
            self._test_permission.stored_permission
        )

        self.grant_access(
            obj=self._test_object, permission=permission_acl_view
        )

        self._clear_events()

        response = self._request_test_acl_permission_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['pk'],
            self._test_permission.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_permission_remove_api_view_no_permission(self):
        self._create_test_acl()
        self._test_acl.permissions.add(
            self._test_permission.stored_permission
        )

        acl_permission_count = self._test_acl.permissions.count()

        self._clear_events()

        response = self._request_test_acl_permission_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_acl.permissions.count(), acl_permission_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_acl_permission_remove_api_view_with_access(self):
        self._create_test_acl()
        self._test_acl.permissions.add(
            self._test_permission.stored_permission
        )

        self.grant_access(
            obj=self._test_object, permission=permission_acl_edit
        )

        acl_permission_count = self._test_acl.permissions.count()

        self._clear_events()

        response = self._request_test_acl_permission_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self._test_acl.permissions.count(), acl_permission_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_acl)
        self.assertEqual(events[0].verb, event_acl_edited.id)


class ClassPermissionAPIViewTestCase(
    ACLTestMixin, ACLAPIViewTestMixin, BaseAPITestCase
):
    auto_create_test_object = True

    def test_class_permission_list_api_view(self):
        class_permissions = [
            permission.pk for permission in ModelPermission.get_for_class(
                klass=self._test_object_content_type.model_class()
            )
        ]

        response = self._request_test_class_permission_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response_permissions = [
            permission['pk'] for permission in response.data['results']
        ]

        self.assertEqual(class_permissions, response_permissions)



from pathlib import Path

from ..events import (
    event_document_file_created, event_document_file_deleted,
    event_document_file_edited, event_document_version_created,
    event_document_version_edited, event_document_version_page_created,
    event_document_version_page_deleted
)

from .base import GenericDocumentTestCase
from .literals import TEST_DOCUMENT_SMALL_CHECKSUM
from .mixins.document_file_mixins import DocumentFileTestMixin


class DocumentFileTestCase(DocumentFileTestMixin, GenericDocumentTestCase):
    def test_file_create(self):
        document_file_count = self._test_document.files.count()

        self._clear_events()

        self._upload_test_document_file(user=self._test_case_user)

        self.assertEqual(
            self._test_document.files.count(), document_file_count + 1
        )
        self.assertEqual(
            self._test_document.file_latest.checksum,
            TEST_DOCUMENT_SMALL_CHECKSUM
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(
            events[3].action_object, self._test_document_version
        )
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[4].action_object, self._test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version)
        self.assertEqual(events[4].verb, event_document_version_edited.id)

    def test_document_file_delete(self):
        document_file_count = self._test_document.files.count()

        self._clear_events()

        self._test_document_file.delete(user=self._test_case_user)

        self.assertEqual(
            self._test_document.files.count(), document_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version)
        self.assertEqual(
            events[0].verb, event_document_version_page_deleted.id
        )

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document)
        self.assertEqual(
            events[1].verb, event_document_file_deleted.id
        )

    def test_document_file_filename_extraction(self):
        """
        Ensure only the filename is stored and not the entire path of the
        uploaded document file.
        """
        self.assertEqual(
            Path(self._test_document_file.filename).name,
            self._test_document_file.filename
        )

    def test_document_first_file_filename(self):
        """
        Ensure the filename of the first document is the same as the document
        label.
        """
        self.assertEqual(
            self._test_document_file.filename, self._test_document.label
        )

    def test_method_get_absolute_url(self):
        self._clear_events()

        self.assertTrue(
            self._test_document.file_latest.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.document_indexing.models.index_instance_models import IndexInstanceNode
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin
from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .literals import TEST_CHECKOUT_INDEX_NODE_TEMPLATE
from .mixins import DocumentCheckoutTestMixin


class CheckoutIndexingTestCase(
    DocumentCheckoutTestMixin, IndexTemplateTestMixin,
    GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_CHECKOUT_INDEX_NODE_TEMPLATE
    auto_create_test_document_stub = True
    auto_upload_test_document = False

    def test_indexing_no_checkout(self):
        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
            ).exists()
        )

    def test_indexing_document_check_in(self):
        self._check_out_test_document()

        self._check_in_test_document()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
            ).exists()
        )

    def test_indexing_document_check_out(self):
        self._check_out_test_document()

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
            ).exists()
        )



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('storage', '0006_auto_20201024_1550')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='downloadfile',
            options={
                'ordering': ('-datetime',), 'verbose_name': 'Download file',
                'verbose_name_plural': 'Download files'
            }
        )
    ]



import datetime


class ValueTransformationTimezoneMixin:
    def _execute(self):
        value = self._execute_()

        if value is not None:
            pass
            try:
                tzinfo = value.tzinfo
            except AttributeError:
                return value
            else:
                if not tzinfo:
                    value = value.replace(tzinfo=datetime.timezone.utc)

        return value



from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_models import Document
from mayan.apps.views.generics import (
    SingleObjectCreateView, SingleObjectDeleteView, SingleObjectDetailView,
    SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .forms import DocumentCommentDetailForm
from .icons import (
    icon_comment_add, icon_comment_delete, icon_comment_detail,
    icon_comment_edit, icon_comments_for_document
)
from .links import link_comment_add
from .models import Comment
from .permissions import (
    permission_document_comment_create, permission_document_comment_delete,
    permission_document_comment_edit, permission_document_comment_view
)


class DocumentCommentCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_permission = permission_document_comment_create
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    fields = ('text',)
    view_icon = icon_comment_add

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(message='Add comment to document: %s') % self.external_object
        }

    def get_instance_extra_data(self):
        return {
            'document': self.external_object,
            'user': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='comments:comments_for_document', kwargs={
                'document_id': self.kwargs['document_id']
            }
        )

    def get_queryset(self):
        return self.external_object.comments.all()


class DocumentCommentDeleteView(SingleObjectDeleteView):
    object_permission = permission_document_comment_delete
    pk_url_kwarg = 'comment_id'
    view_icon = icon_comment_delete

    def get_extra_context(self):
        return {
            'comment': self.object,
            'document': self.object.document,
            'navigation_object_list': ('document', 'comment'),
            'title': _(message='Delete comment: %s?') % self.object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='comments:comments_for_document', kwargs={
                'document_id': self.object.document_id
            }
        )

    def get_source_queryset(self):
        return Comment.objects.filter(
            document_id__in=Document.valid.values('id')
        )


class DocumentCommentDetailView(SingleObjectDetailView):
    form_class = DocumentCommentDetailForm
    pk_url_kwarg = 'comment_id'
    object_permission = permission_document_comment_view
    view_icon = icon_comment_detail

    def get_extra_context(self):
        return {
            'comment': self.object,
            'document': self.object.document,
            'navigation_object_list': ('document', 'comment'),
            'title': _(message='Details for comment: %s?') % self.object
        }

    def get_source_queryset(self):
        return Comment.objects.filter(
            document_id__in=Document.valid.values('id')
        )


class DocumentCommentEditView(SingleObjectEditView):
    fields = ('text',)
    pk_url_kwarg = 'comment_id'
    object_permission = permission_document_comment_edit
    view_icon = icon_comment_edit

    def get_extra_context(self):
        return {
            'comment': self.object,
            'document': self.object.document,
            'navigation_object_list': ('document', 'comment'),
            'title': _(message='Edit comment: %s?') % self.object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='comments:comments_for_document', kwargs={
                'document_id': self.object.document_id
            }
        )

    def get_source_queryset(self):
        return Comment.objects.filter(
            document_id__in=Document.valid.values('id')
        )


class DocumentCommentListView(ExternalObjectViewMixin, SingleObjectListView):
    external_object_permission = permission_document_comment_view
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    view_icon = icon_comments_for_document

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_comments_for_document,
            'no_results_external_link': link_comment_add.resolve(
                RequestContext(
                    dict_={'object': self.external_object},
                    request=self.request
                )
            ),
            'no_results_text': _(
                'Document comments are timestamped text entries from users. '
                'They are great for collaboration.'
            ),
            'no_results_title': _(message='There are no comments'),
            'object': self.external_object,
            'title': _(message='Comments for document: %s') % self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.comments.all()



from django.db import models


class ValidSignatureCaptureManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(
            document__in_trash=False
        )



import logging

from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.generics import (
    MultipleObjectFormActionView, SingleObjectCreateView,
    SingleObjectDeleteView, SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .forms import CabinetListForm
from .icons import (
    icon_cabinet, icon_cabinet_child_add, icon_cabinet_create,
    icon_cabinet_delete, icon_cabinet_detail, icon_cabinet_edit,
    icon_cabinet_list, icon_document_cabinet_add,
    icon_document_cabinet_list, icon_document_cabinet_remove
)
from .links import (
    link_cabinet_child_add, link_cabinet_create, link_document_cabinet_add
)
from .models import Cabinet
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_remove_document, permission_cabinet_view
)
from .widgets import jstree_data

logger = logging.getLogger(name=__name__)


class CabinetCreateView(SingleObjectCreateView):
    fields = ('label',)
    model = Cabinet
    post_action_redirect = reverse_lazy(viewname='cabinets:cabinet_list')
    view_icon = icon_cabinet_create
    view_permission = permission_cabinet_create

    def get_extra_context(self):
        return {
            'title': _(message='Create cabinet')
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class CabinetChildAddView(ExternalObjectViewMixin, SingleObjectCreateView):
    fields = ('label',)
    external_object_class = Cabinet
    external_object_permission = permission_cabinet_create
    external_object_pk_url_kwarg = 'cabinet_id'
    view_icon = icon_cabinet_child_add

    def get_extra_context(self):
        return {
            'title': _(
                'Add new level to: %s'
            ) % self.external_object.get_full_path(),
            'object': self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'parent': self.external_object
        }

    def get_queryset(self):
        return self.external_object.get_descendants()


class CabinetDeleteView(SingleObjectDeleteView):
    model = Cabinet
    object_permission = permission_cabinet_delete
    post_action_redirect = reverse_lazy(viewname='cabinets:cabinet_list')
    pk_url_kwarg = 'cabinet_id'
    view_icon = icon_cabinet_delete

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Delete the cabinet: %s?') % self.object
        }


class CabinetDetailView(ExternalObjectViewMixin, DocumentListView):
    external_object_class = Cabinet
    external_object_permission = permission_cabinet_view
    external_object_pk_url_kwarg = 'cabinet_id'
    template_name = 'cabinets/cabinet_details.html'
    view_icon = icon_cabinet_detail

    def get_document_queryset(self):
        return self.external_object.get_documents(
            permission=permission_document_view, user=self.request.user
        )

    def get_extra_context(self, **kwargs):
        context = super().get_extra_context(**kwargs)

        context.update(
            {
                'column_class': 'col-xs-12 col-sm-6 col-md-4 col-lg-3',
                'hide_links': True,
                'jstree_data': '\n'.join(
                    jstree_data(
                        node=self.external_object.get_root(),
                        selected_node=self.external_object
                    )
                ),
                'list_as_items': True,
                'no_results_icon': icon_cabinet,
                'no_results_main_link': link_cabinet_child_add.resolve(
                    context=RequestContext(
                        dict_={
                            'object': self.external_object
                        }, request=self.request
                    )
                ),
                'no_results_text': _(
                    'Cabinet levels can contain documents or other '
                    'cabinet sub levels. To add documents to a cabinet, '
                    'select the cabinet view of a document view.'
                ),
                'no_results_title': _(message='This cabinet level is empty'),
                'object': self.external_object,
                'title': _(
                    'Details of cabinet: %s'
                ) % self.external_object.get_full_path()
            }
        )

        return context


class CabinetEditView(SingleObjectEditView):
    fields = ('label',)
    model = Cabinet
    object_permission = permission_cabinet_edit
    post_action_redirect = reverse_lazy(viewname='cabinets:cabinet_list')
    pk_url_kwarg = 'cabinet_id'
    view_icon = icon_cabinet_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Edit cabinet: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class CabinetListView(SingleObjectListView):
    object_permission = permission_cabinet_view
    view_icon = icon_cabinet_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'title': _(message='Cabinets'),
            'no_results_icon': icon_cabinet,
            'no_results_main_link': link_cabinet_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Cabinets are a multi-level method to organize '
                'documents. Each cabinet can contain documents as '
                'well as other sub level cabinets.'
            ),
            'no_results_title': _(message='No cabinets available')
        }

    def get_source_queryset(self):
        return Cabinet.objects.root_nodes()


class DocumentCabinetAddView(MultipleObjectFormActionView):
    form_class = CabinetListForm
    object_permission = permission_cabinet_add_document
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message_single = _(
        'Document "%(object)s" added to cabinets successfully.'
    )
    success_message_singular = _(
        '%(count)d document added to cabinets successfully.'
    )
    success_message_plural = _(
        '%(count)d documents added to cabinets successfully.'
    )
    title_plural = _(message='Add %(count)d documents to cabinets.')
    title_single = _(message='Add document "%(object)s" to cabinets.')
    title_singular = _(message='Add %(count)d document to cabinets.')
    view_icon = icon_document_cabinet_add

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def get_form_extra_kwargs(self):
        kwargs = {
            'help_text': _(
                'Cabinets to which the selected documents will be added.'
            ),
            'permission': permission_cabinet_add_document,
            'queryset': Cabinet.objects.all(),
            'user': self.request.user
        }

        if self.object_list.count() == 1:
            kwargs.update(
                {
                    'queryset': Cabinet.objects.exclude(
                        pk__in=self.object_list.first().cabinets.all()
                    )
                }
            )

        return kwargs

    def object_action(self, form, instance):
        for cabinet in form.cleaned_data['cabinets']:
            AccessControlList.objects.check_access(
                obj=cabinet, permission=permission_cabinet_add_document,
                user=self.request.user
            )

            cabinet.document_add(document=instance, user=self.request.user)


class DocumentCabinetListView(ExternalObjectViewMixin, CabinetListView):
    external_object_permission = permission_cabinet_view
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    view_icon = icon_document_cabinet_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'no_results_icon': icon_cabinet,
            'no_results_main_link': link_document_cabinet_add.resolve(
                context=RequestContext(
                    dict_={
                        'object': self.external_object
                    }, request=self.request
                )
            ),
            'no_results_text': _(
                'Documents can be added to many cabinets.'
            ),
            'no_results_title': _(
                'This document is not in any cabinet'
            ),
            'object': self.external_object,
            'title': _(
                'Cabinets containing document: %s'
            ) % self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.get_cabinets(
            permission=permission_cabinet_view, user=self.request.user
        )


class DocumentCabinetRemoveView(MultipleObjectFormActionView):
    form_class = CabinetListForm
    object_permission = permission_cabinet_remove_document
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message_single = _(
        'Document "%(object)s" removed from cabinets successfully.'
    )
    success_message_singular = _(
        '%(count)d document removed from cabinets successfully.'
    )
    success_message_plural = _(
        '%(count)d documents removed from cabinets successfully.'
    )
    title_plural = _(message='Remove %(count)d documents from cabinets.')
    title_single = _(message='Remove document "%(object)s" from cabinets.')
    title_singular = _(message='Remove %(count)d document from cabinets.')
    view_icon = icon_document_cabinet_remove

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def get_form_extra_kwargs(self):
        kwargs = {
            'help_text': _(
                'Cabinets from which the selected documents will be removed.'
            ),
            'permission': permission_cabinet_remove_document,
            'queryset': Cabinet.objects.all(),
            'user': self.request.user
        }

        if self.object_list.count() == 1:
            kwargs.update(
                {
                    'queryset': self.object_list.first().cabinets.all()
                }
            )

        return kwargs

    def object_action(self, form, instance):
        for cabinet in form.cleaned_data['cabinets']:
            AccessControlList.objects.check_access(
                obj=cabinet, permission=permission_cabinet_remove_document,
                user=self.request.user
            )

            cabinet.document_remove(
                document=instance, user=self.request.user
            )



from itertools import islice
import logging
import os

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from mayan.apps.converter.exceptions import AppImageError
from mayan.apps.databases.classes import ModelQueryFields
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter
from mayan.apps.templating.classes import Template

from ..events import (
    event_document_version_page_created, event_document_version_edited
)
from ..literals import (
    DOCUMENT_VERSION_PAGE_CREATE_BATCH_SIZE, IMAGE_ERROR_NO_VERSION_PAGES,
    STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE
)
from ..signals import signal_post_document_version_remap

logger = logging.getLogger(name=__name__)


class DocumentVersionBusinessLogicMixin:
    @staticmethod
    def annotate_content_object_list(
        content_object_list, start_page_number=None
    ):
        def content_object_to_dictionary(entry):
            # Argument order based on the return value of enumerate.
            return {
                'content_object': entry[1],
                'page_number': entry[0]
            }

        return map(
            content_object_to_dictionary, enumerate(
                iterable=content_object_list or (),
                start=start_page_number or 1
            )
        )

    def active_set(self, save=True):
        with transaction.atomic():
            self.document.versions.exclude(pk=self.pk).update(active=False)

            self.active = True

            if save:
                self.save(
                    update_fields=('active',)
                )

            self.document.version_active = self
            self.document._event_ignore = True
            self.document.save(
                update_fields=('version_active',)
            )

    @cached_property
    def cache(self):
        Cache = apps.get_model(app_label='file_caching', model_name='Cache')

        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='version-{}'.format(self.uuid)
        )
        return partition

    def get_absolute_api_url(self):
        return reverse(
            viewname='rest_api:documentversion-detail', kwargs={
                'document_id': self.document_id,
                'document_version_id': self.pk
            }
        )

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        first_page = self.pages.first()
        if first_page:
            return first_page.get_api_image_url(
                maximum_layer_order=maximum_layer_order,
                transformation_instance_list=transformation_instance_list,
                user=user
            )
        else:
            raise AppImageError(error_name=IMAGE_ERROR_NO_VERSION_PAGES)

    def get_cache_partitions(self):
        result = [self.cache_partition]
        for page in self.version_pages.all():
            result.append(page.cache_partition)

        return result

    def get_label(self, preserve_extension=False):
        if preserve_extension:
            filename, extension = os.path.splitext(self.document.label)
            return Template(
                template_string='{{ filename }} ({{ instance.timestamp }}){{ extension }}'
            ).render(
                context={
                    'extension': extension,
                    'filename': filename,
                    'instance': self
                }
            )
        else:
            return Template(
                template_string='{{ instance.document }} ({{ instance.timestamp }})'
            ).render(
                context={'instance': self}
            )
    get_label.short_description = _(message='Label')

    def get_source_content_object_dictionary_list(self):
        DocumentFilePage = apps.get_model(
            app_label='documents', model_name='DocumentFilePage'
        )

        content_object_dictionary_list = []

        document_file_page_content_type = ContentType.objects.get_for_model(
            model=DocumentFilePage
        )

        for document_file in self.document.files.all():
            for document_file_page in document_file.pages.all():
                content_object_dictionary_list.append(
                    {
                        'content_type': document_file_page_content_type,
                        'object_id': document_file_page.pk
                    }
                )

        return content_object_dictionary_list

    @property
    def is_in_trash(self):
        return self.document.is_in_trash

    @property
    def page_content_objects(self):
        result = []
        for page in self.pages.all():
            result.append(page.content_object)

        return result

    @property
    def pages(self):
        DocumentVersionPage = apps.get_model(
            app_label='documents', model_name='DocumentVersionPage'
        )

        queryset = ModelQueryFields.get(
            model=DocumentVersionPage
        ).get_queryset()
        return queryset.filter(
            pk__in=self.version_pages.all()
        )

    def pages_append_all(self, user=None):
        """
        Append the pages of all document files.
        """
        DocumentFilePage = apps.get_model(
            app_label='documents', model_name='DocumentFilePage'
        )
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        document_file_pages = DocumentFilePage.objects.filter(
            document_file__document=self.document
        ).order_by('document_file__timestamp', 'page_number')

        annotated_content_object_list = DocumentVersion.annotate_content_object_list(
            content_object_list=list(document_file_pages)
        )
        return self.pages_remap(
            annotated_content_object_list=annotated_content_object_list,
            user=user
        )

    @property
    def pages_first(self):
        return self.pages.first()

    @method_event(
        action_object='document',
        event_manager_class=EventManagerMethodAfter,
        event=event_document_version_edited,
        target='self'
    )
    def pages_remap(self, annotated_content_object_list=None, user=None):
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )
        DocumentVersionPage = apps.get_model(
            app_label='documents', model_name='DocumentVersionPage'
        )

        self._event_actor = user

        for page in self.pages.all():
            page._event_actor = user
            page.delete()

        if not annotated_content_object_list:
            annotated_content_object_list = ()

        document_version_pages = (
            DocumentVersionPage(
                document_version=self,
                content_object=content_object_entry['content_object'],
                page_number=content_object_entry['page_number']
            ) for content_object_entry in annotated_content_object_list
        )

        while True:
            batch = list(
                islice(
                    document_version_pages,
                    DOCUMENT_VERSION_PAGE_CREATE_BATCH_SIZE
                )
            )

            if not batch:
                break

            DocumentVersionPage.objects.bulk_create(
                batch_size=DOCUMENT_VERSION_PAGE_CREATE_BATCH_SIZE,
                objs=batch
            )

        for page in self.pages.all().only('pk'):
            event_document_version_page_created.commit(
                action_object=self, actor=user, target=page
            )

        signal_post_document_version_remap.send(
            instance=self, sender=DocumentVersion
        )

    def pages_reset(self, document_file=None, user=None):
        """
        Remove all page mappings and recreate them to be a 1 to 1 match
        to the latest document file or the document file supplied.
        """
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        latest_file = document_file or self.document.file_latest

        if latest_file:
            content_object_list = list(
                latest_file.pages.all()
            )
        else:
            content_object_list = None

        annotated_content_object_list = DocumentVersion.annotate_content_object_list(
            content_object_list=content_object_list
        )
        return self.pages_remap(
            annotated_content_object_list=annotated_content_object_list,
            user=user
        )

    @property
    def uuid(self):
        # Make cache UUID a mix of document UUID, file ID.
        return '{}-{}'.format(self.document.uuid, self.pk)



from .base import GenericDocumentTestCase
from .literals import TEST_FILE_MULTI_PAGE_TIFF_FILENAME, TEST_FILE_OFFICE


class OfficeTextDocumentTestCase(GenericDocumentTestCase):
    _test_document_filename = TEST_FILE_OFFICE

    def test_document_creation(self):
        self.assertEqual(
            self._test_document.file_latest.mimetype, 'application/msword'
        )
        self.assertEqual(
            self._test_document.file_latest.encoding, 'binary'
        )
        self.assertEqual(
            self._test_document.file_latest.checksum,
            '03a7e9071d2c6ae05a6588acd7dff1d890fac2772cf61abd470c9ffa6ef71f03'
        )
        self.assertEqual(self._test_document.pages.count(), 2)


class MultiPageTiffTestCase(GenericDocumentTestCase):
    _test_document_filename = TEST_FILE_MULTI_PAGE_TIFF_FILENAME

    def test_document_creation(self):
        self.assertEqual(self._test_document.file_latest.mimetype, 'image/tiff')
        self.assertEqual(self._test_document.file_latest.encoding, 'binary')
        self.assertEqual(
            self._test_document.file_latest.checksum,
            '40adaa9d658b65c70a7f002dfe084a8354bb77c0dfbf1993e31fb024a285fb1d'
        )
        self.assertEqual(self._test_document.pages.count(), 2)



from django.apps import apps
from django.contrib.auth import get_user_model
from django.shortcuts import reverse

from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .events import (
    event_group_created, event_group_edited, event_user_created,
    event_user_edited
)
from .permissions import permission_group_view, permission_user_view
from .querysets import get_user_queryset


def get_method_group_init():
    Group = apps.get_model(app_label='auth', model_name='Group')
    method_original = Group.__init__

    def method_init(self, *args, **kwargs):
        _instance_extra_data = kwargs.pop(
            '_instance_extra_data', {}
        )
        result = method_original(self, *args, **kwargs)
        for key, value in _instance_extra_data.items():
            setattr(self, key, value)

        return result

    return method_init


def get_method_group_save():
    Group = apps.get_model(app_label='auth', model_name='Group')
    group_save_original = Group.save

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_group_created,
            'target': 'self',
        },
        edited={
            'event': event_group_edited,
            'target': 'self',
        }
    )
    def method_group_save(self, *args, **kwargs):
        group_save_original(self, *args, **kwargs)

    return method_group_save


def method_group_get_absolute_url(self):
    return reverse(
        viewname='user_management:group_detail', kwargs={'group_id': self.pk}
    )


def method_group_get_users(self, user, permission=permission_user_view):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )

    return AccessControlList.objects.restrict_queryset(
        permission=permission, queryset=get_user_queryset().filter(
            id__in=self.user_set.all()
        ), user=user
    )


def method_group_users_add(self, queryset, user=None):
    for model_instance in queryset:
        self.user_set.add(model_instance)
        event_group_edited.commit(
            action_object=model_instance, actor=user, target=self
        )


def method_group_users_remove(self, queryset, user=None):
    for model_instance in queryset:
        self.user_set.remove(model_instance)
        event_group_edited.commit(
            action_object=model_instance, actor=user, target=self
        )


def get_method_user_init():
    User = get_user_model()
    method_original = User.__init__

    def method_init(self, *args, **kwargs):
        _instance_extra_data = kwargs.pop(
            '_instance_extra_data', {}
        )
        result = method_original(self, *args, **kwargs)
        for key, value in _instance_extra_data.items():
            setattr(self, key, value)

        return result

    return method_init


def method_user_get_absolute_url(self):
    return reverse(
        viewname='user_management:user_details', kwargs={'user_id': self.pk}
    )


def method_user_get_absolute_api_url(self):
    return reverse(
        viewname='rest_api:user-detail', kwargs={'user_id': self.pk}
    )


def method_user_get_groups(self, user, permission=permission_group_view):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )

    return AccessControlList.objects.restrict_queryset(
        permission=permission, queryset=self.groups.all(), user=user
    )


def method_user_groups_add(self, queryset, user=None):
    for model_instance in queryset:
        self.groups.add(model_instance)
        event_user_edited.commit(
            action_object=model_instance, actor=user, target=self
        )


def method_user_groups_remove(self, queryset, user=None):
    for model_instance in queryset:
        self.groups.remove(model_instance)
        event_user_edited.commit(
            action_object=model_instance, actor=user, target=self
        )


def get_method_user_save():
    user_save_original = get_user_model().save

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_user_created,
            'target': 'self'
        },
        edited={
            'event': event_user_edited,
            'target': 'self'
        }
    )
    def method_user_save(self, *args, **kwargs):
        user_save_original(self, *args, **kwargs)

    return method_user_save



import logging

from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_object, menu_return, menu_secondary, menu_setup
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn

from .classes import CredentialBackend
from .events import event_credential_edited, event_credential_used
from .links import (
    link_credential_backend_selection, link_credential_delete,
    link_credential_edit, link_credential_list, link_credential_setup
)
from .permissions import (
    permission_credential_delete, permission_credential_edit,
    permission_credential_use, permission_credential_view
)

logger = logging.getLogger(name=__name__)


class CredentialsApp(MayanAppConfig):
    app_namespace = 'credentials'
    app_url = 'credentials'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.credentials'
    verbose_name = _(message='Credentials')

    def ready(self):
        super().ready()

        CredentialBackend.load_modules()

        StoredCredential = self.get_model(model_name='StoredCredential')

        EventModelRegistry.register(model=StoredCredential)

        ModelEventType.register(
            model=StoredCredential, event_types=(
                event_credential_edited, event_credential_used
            )
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=StoredCredential
        )
        SourceColumn(
            attribute='internal_name', include_label=True, is_sortable=True,
            source=StoredCredential
        )
        SourceColumn(
            attribute='get_backend_class_label', include_label=True,
            source=StoredCredential
        )

        ModelPermission.register(
            model=StoredCredential, permissions=(
                permission_credential_delete, permission_credential_edit,
                permission_credential_view, permission_credential_use
            )
        )

        menu_object.bind_links(
            links=(link_credential_delete, link_credential_edit),
            sources=(StoredCredential,)
        )
        menu_return.bind_links(
            links=(link_credential_list,), sources=(
                StoredCredential,
                'credentials:stored_credential_backend_selection',
                'credentials:stored_credential_create',
                'credentials:stored_credential_list'
            )
        )
        menu_secondary.bind_links(
            links=(link_credential_backend_selection,), sources=(
                StoredCredential,
                'credentials:stored_credential_backend_selection',
                'credentials:stored_credential_create',
                'credentials:stored_credential_list'
            )
        )

        menu_setup.bind_links(
            links=(link_credential_setup,)
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='OCR'), name='ocr'
)

event_ocr_document_version_content_deleted = namespace.add_event_type(
    label=_(message='Document version OCR content deleted'),
    name='document_content_deleted'
)
event_ocr_document_version_page_content_edited = namespace.add_event_type(
    label=_(message='Document version page OCR content edited'),
    name='document_version_page_content_edited'
)
event_ocr_document_version_submitted = namespace.add_event_type(
    label=_(message='Document version submitted for OCR'),
    name='document_version_submit'
)
event_ocr_document_version_finished = namespace.add_event_type(
    label=_(message='Document version OCR finished'),
    name='document_version_finish'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_caching', '0002_auto_20190729_0236')
    ]

    operations = [
        migrations.AlterField(
            model_name='cache',
            name='name',
            field=models.CharField(
                db_index=True, help_text='Internal name of the cache.',
                max_length=128, unique=True, verbose_name='Name'
            )
        )
    ]



import uuid

from django.db import models, transaction
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.signals import signal_mayan_pre_save
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from ..events import (
    event_document_created, event_document_edited, event_document_trashed,
    event_trashed_document_deleted
)
from ..literals import DEFAULT_LANGUAGE
from ..managers import (
    DocumentManager, TrashCanManager, ValidDocumentManager,
    ValidRecentlyCreatedDocumentManager
)
from ..settings import setting_language

from .document_model_mixins import DocumentBusinessLogicMixin
from .document_type_models import DocumentType
from .model_mixins import HooksModelMixin

__all__ = ('Document', 'DocumentSearchResult',)


class Document(
    DocumentBusinessLogicMixin, ExtraDataModelMixin, HooksModelMixin,
    models.Model
):
    """
    Defines a single document with it's fields and properties
    Fields:
    * uuid - UUID of a document, universally Unique ID. An unique identifier
    generated for each document. No two documents can ever have the same UUID.
    This ID is generated automatically.
    """
    _hooks_pre_create = []

    file_latest = models.OneToOneField(
        blank=True, null=True, on_delete=models.SET_NULL, to='DocumentFile',
        related_name='document_latest', verbose_name=_(
            'Latest document file'
        )
    )
    uuid = models.UUIDField(
        default=uuid.uuid4, editable=False, help_text=_(
            'UUID of a document, universally Unique ID. An unique '
            'identifier generated for each document.'
        ), verbose_name=_(message='UUID')
    )
    document_type = models.ForeignKey(
        help_text=_(message='The document type of the document.'),
        on_delete=models.CASCADE, related_name='documents', to=DocumentType,
        verbose_name=_(message='Document type')
    )
    label = models.CharField(
        blank=True, db_index=True, default='', max_length=255,
        help_text=_(
            'A short text identifying the document. By default, will be '
            'set to the filename of the first file uploaded to the document.'
        ),
        verbose_name=_(message='Label')
    )
    description = models.TextField(
        blank=True, default='', help_text=_(
            'An optional short text describing a document.'
        ), verbose_name=_(message='Description')
    )
    datetime_created = models.DateTimeField(
        auto_now_add=True, db_index=True, help_text=_(
            'The date and time of the document creation.'
        ), verbose_name=_(message='Created')
    )
    language = models.CharField(
        blank=True, default=DEFAULT_LANGUAGE, help_text=_(
            'The primary language in the document.'
        ), max_length=8, verbose_name=_(message='Language')
    )
    in_trash = models.BooleanField(
        db_index=True, default=False, help_text=_(
            'Whether or not this document is in the trash.'
        ), editable=False, verbose_name=_(message='In trash?')
    )
    trashed_date_time = models.DateTimeField(
        blank=True, editable=True, help_text=_(
            'The server date and time when the document was moved to the '
            'trash.'
        ), null=True, verbose_name=_(message='Date and time trashed')
    )
    is_stub = models.BooleanField(
        db_index=True, default=True, editable=False, help_text=_(
            'A document stub is a document with an entry on the database '
            'but no file uploaded. This could be an interrupted upload or '
            'a deferred upload via the API.'
        ), verbose_name=_(message='Is stub?')
    )
    version_active = models.OneToOneField(
        blank=True, null=True, on_delete=models.SET_NULL,
        to='DocumentVersion', related_name='document_active', verbose_name=_(
            'Active document version'
        )
    )

    objects = DocumentManager()
    trash = TrashCanManager()
    valid = ValidDocumentManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _(message='Document')
        verbose_name_plural = _(message='Documents')

    def __str__(self):
        return self.get_label()

    def delete(self, *args, **kwargs):
        to_trash = kwargs.pop('to_trash', True)
        user = self.__dict__.pop('_event_actor', None)

        if not self.in_trash and to_trash:
            self.in_trash = True
            self.trashed_date_time = now()
            with transaction.atomic():
                self._event_ignore = True
                self.save(
                    update_fields=('in_trash', 'trashed_date_time')
                )

            event_document_trashed.commit(actor=user, target=self)
        else:
            with transaction.atomic():
                for document_file in self.files.all():
                    document_file.delete()

                super().delete(*args, **kwargs)

            event_trashed_document_deleted.commit(
                actor=user, target=self.document_type
            )

    def get_absolute_url(self):
        return reverse(
            viewname='documents:document_preview', kwargs={
                'document_id': self.pk
            }
        )

    def natural_key(self):
        return (self.uuid,)
    natural_key.dependencies = ['documents.DocumentType']

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_document_created,
            'action_object': 'document_type',
            'target': 'self'
        },
        edited={
            'event': event_document_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        user = self.__dict__.pop('_event_actor', None)
        new_document = not self.pk

        self.description = self.description or ''
        self.label = self.label or ''
        self.language = self.language or setting_language.value

        signal_mayan_pre_save.send(
            instance=self, sender=Document, user=user
        )

        super().save(*args, **kwargs)

        if new_document:
            if user:
                self.add_as_recent_document_for_user(user=user)


class DocumentSearchResult(Document):
    class Meta:
        proxy = True


class RecentlyCreatedDocument(Document):
    objects = models.Manager()
    valid = ValidRecentlyCreatedDocumentManager()

    class Meta:
        proxy = True



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_asset_created, event_asset_edited
from ..models import Asset
from ..permissions import (
    permission_asset_create, permission_asset_delete, permission_asset_edit,
    permission_asset_view
)

from .mixins import AssetAPIViewTestMixin


class AssetAPIViewTestCase(AssetAPIViewTestMixin, BaseAPITestCase):
    def test_asset_create_api_view_no_permission(self):
        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Asset.objects.count(), asset_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_asset_create)

        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Asset.objects.count(), asset_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_asset)
        self.assertEqual(events[0].verb, event_asset_created.id)

    def test_asset_delete_api_view_no_permission(self):
        self._create_test_asset()

        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Asset.objects.count(), asset_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_delete_api_view_with_access(self):
        self._create_test_asset()

        self.grant_access(obj=self._test_asset, permission=permission_asset_delete)

        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Asset.objects.count(), asset_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_detail_api_view_no_permission(self):
        self._create_test_asset()

        self._clear_events()

        response = self._request_test_asset_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_detail_api_view_with_access(self):
        self._create_test_asset()
        self.grant_access(
            obj=self._test_asset, permission=permission_asset_view
        )

        self._clear_events()

        response = self._request_test_asset_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data['label'], self._test_asset.label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_edit_api_view_via_patch_no_permission(self):
        self._create_test_asset()

        asset_label = self._test_asset.label

        self._clear_events()

        response = self._request_test_asset_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_asset.refresh_from_db()
        self.assertEqual(self._test_asset.label, asset_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_edit_api_view_via_patch_with_access(self):
        self._create_test_asset()

        self.grant_access(obj=self._test_asset, permission=permission_asset_edit)

        asset_label = self._test_asset.label

        self._clear_events()

        response = self._request_test_asset_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_asset.refresh_from_db()
        self.assertNotEqual(self._test_asset.label, asset_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_asset)
        self.assertEqual(events[0].verb, event_asset_edited.id)

    def test_asset_edit_api_view_via_put_no_permission(self):
        self._create_test_asset()

        asset_label = self._test_asset.label

        self._clear_events()

        response = self._request_test_asset_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_asset.refresh_from_db()
        self.assertEqual(self._test_asset.label, asset_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_edit_api_view_via_put_with_access(self):
        self._create_test_asset()

        self.grant_access(obj=self._test_asset, permission=permission_asset_edit)

        asset_label = self._test_asset.label

        self._clear_events()

        response = self._request_test_asset_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_asset.refresh_from_db()
        self.assertNotEqual(self._test_asset.label, asset_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_asset)
        self.assertEqual(events[0].verb, event_asset_edited.id)

    def test_asset_list_api_view_no_permission(self):
        self._create_test_asset()

        self._clear_events()

        response = self._request_test_asset_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_list_api_view_with_access(self):
        self._create_test_asset()
        self.grant_access(
            obj=self._test_asset, permission=permission_asset_view
        )

        self._clear_events()

        response = self._request_test_asset_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['label'],
            self._test_asset.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.documents.literals import DEFAULT_DOCUMENT_FILE_ACTION_NAME
from mayan.apps.source_compressed.source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_NEVER
from mayan.apps.source_stored_files.tests.literals import (
    TEST_STAGING_PREVIEW_MAX_SIZE, TEST_STAGING_PREVIEW_WIDTH
)
from mayan.apps.source_stored_files.tests.mixins import SourceTestMixinStoredFile
from mayan.apps.sources.tests.literals import (
    TEST_CASE_ACTION_NAME_SOURCE_CREATE, TEST_CASE_INTERFACE_NAME_MODEL,
    TEST_CASE_INTERFACE_NAME_REST_API, TEST_CASE_INTERFACE_NAME_VIEW
)

from .literals import TEST_SOURCE_BACKEND_PATH_STAGING_STORAGE


class StagingStorageSourceTestMixin(SourceTestMixinStoredFile):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_STAGING_STORAGE

    def _get_test_source_backend_data(self, action_name, interface_name):
        result = super()._get_test_source_backend_data(
            action_name=action_name, interface_name=interface_name
        )

        if action_name == TEST_CASE_ACTION_NAME_SOURCE_CREATE:
            result.update(
                {
                    'delete_after_upload': True,
                    'preview_max_size': TEST_STAGING_PREVIEW_MAX_SIZE,
                    'preview_width': TEST_STAGING_PREVIEW_WIDTH,
                    'storage_backend': 'django.core.files.storage.FileSystemStorage',
                    'storage_backend_arguments': '{{\'location\': \'{}\'}}'.format(
                        self._test_source_temporary_folder
                    ),
                    'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER
                }
            )
        elif action_name == 'document_file_upload':
            if interface_name == TEST_CASE_INTERFACE_NAME_MODEL:
                result.update(
                    {
                        'document': self._test_document,
                        'file_identifier': self._test_source_stored_file.encoded_filename
                    }
                )
            elif interface_name == TEST_CASE_INTERFACE_NAME_REST_API:
                result.update(
                    {
                        'document_id': self._test_document.pk,
                        'encoded_filename': self._test_source_stored_file.encoded_filename
                    }
                )
            elif interface_name == TEST_CASE_INTERFACE_NAME_VIEW:
                result.update(
                    {
                        'source-stored_source_file_id': self._test_source_stored_file.encoded_filename,
                        'document-action_name': DEFAULT_DOCUMENT_FILE_ACTION_NAME
                    }
                )
        elif action_name == 'document_upload':
            if self._test_source_stored_file:
                file_identifier = self._test_source_stored_file.encoded_filename
            else:
                file_identifier = None

            if interface_name == TEST_CASE_INTERFACE_NAME_MODEL:
                result.update(
                    {
                        'document_type': self._test_document_type,
                        'file_identifier': file_identifier
                    }
                )
            elif interface_name == TEST_CASE_INTERFACE_NAME_REST_API:
                result.update(
                    {
                        'document_type_id': self._test_document_type.pk,
                        'encoded_filename': file_identifier
                    }
                )
            elif interface_name == TEST_CASE_INTERFACE_NAME_VIEW:
                result['source-stored_source_file_id'] = file_identifier
        elif action_name == 'file_delete':
            if interface_name == TEST_CASE_INTERFACE_NAME_MODEL:
                result['encoded_filename'] = self._test_source_stored_file.encoded_filename
            elif interface_name == TEST_CASE_INTERFACE_NAME_REST_API:
                result['encoded_filename'] = self._test_source_stored_file.encoded_filename
            elif interface_name == TEST_CASE_INTERFACE_NAME_VIEW:
                result['encoded_filename'] = self._test_source_stored_file.encoded_filename
        elif action_name == 'file_image':
            if interface_name == TEST_CASE_INTERFACE_NAME_MODEL:
                result['encoded_filename'] = self._test_source_stored_file.encoded_filename
            elif interface_name == TEST_CASE_INTERFACE_NAME_REST_API:
                result['encoded_filename'] = self._test_source_stored_file.encoded_filename

        return result



import uuid

TEST_OBJECT_BOOLEAN_VALUE = True
TEST_OBJECT_CHAR_VALUE = 'test char value'
TEST_OBJECT_EMAIL_VALUE = 'user@example.org'
TEST_OBJECT_INTEGER_VALUE = 2**14
TEST_OBJECT_POSITIVE_INTEGER_VALUE = 2**15
TEST_OBJECT_TEXT_VALUE = 'test text value'
TEST_OBJECT_UUID_VALUE = uuid.uuid4()

TEST_QUERY_TYPE_ALIAS_INVALID = '|||'

TEST_SEARCH_OBJECT_TERM = 'document'
TEST_SEARCH_OBJECT_TERMS = ('document', 'stub')

# General

TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER = '0'
TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER_INVALID = 'INVALID'

# Field entry

TEST_SCOPED_QUERY_ENTRY_FIELD_NAME = 'label'
TEST_SCOPED_QUERY_ENTRY_FIELD_NAME_INVALID = 'INVALID'
TEST_SCOPED_QUERY_ENTRY_VALUE = 'TEST_VALUE'

# Operator entry

TEST_SCOPED_QUERY_ENTRY_OPERATOR_SCOPE_IDENTIFIER = '3'
TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERAND_LIST = ('0', '1')
TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERATOR_TEXT = 'AND'

# Result entry

TEST_SCOPED_QUERY_RESULT_SCOPE_IDENTIFIER = '0'



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0088_populate_version_active')
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttype',
            name='delete_time_unit',
            field=models.CharField(
                blank=True, choices=[
                    ('days', 'Days'), ('hours', 'Hours'),
                    ('minutes', 'Minutes')
                ], default='days', max_length=12, null=True,
                verbose_name='Delete time unit'
            )
        ),
        migrations.AlterField(
            model_name='documenttype',
            name='trash_time_unit',
            field=models.CharField(
                blank=True, choices=[
                    ('days', 'Days'), ('hours', 'Hours'),
                    ('minutes', 'Minutes')
                ], max_length=12, null=True, verbose_name='Trash time unit'
            )
        )
    ]



from django.apps import apps
from django.db.models.signals import post_save, pre_delete
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_list_facet, menu_main, menu_multi_item, menu_object, menu_secondary
)
from mayan.apps.databases.classes import ModelFieldRelated, ModelQueryFields
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField

from .column_widgets import DocumentTagWidget
from .events import event_tag_attached, event_tag_edited, event_tag_removed
from .handlers import handler_index_document, handler_tag_pre_delete
from .links import (
    link_document_multiple_tag_multiple_attach,
    link_document_multiple_tag_multiple_remove, link_document_tag_list,
    link_document_tag_multiple_attach, link_document_tag_multiple_remove,
    link_tag_create, link_tag_document_list, link_tag_edit, link_tag_list,
    link_tag_multiple_delete, link_tag_single_delete
)
from .menus import menu_tags
from .methods import method_document_get_tags
from .permissions import (
    permission_tag_attach, permission_tag_delete, permission_tag_edit,
    permission_tag_remove, permission_tag_view
)


class TagsApp(MayanAppConfig):
    app_namespace = 'tags'
    app_url = 'tags'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.tags'
    verbose_name = _(message='Tags')

    def ready(self):
        super().ready()
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        DocumentFileSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentFileSearchResult'
        )
        DocumentFilePageSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentFilePageSearchResult'
        )
        DocumentVersionSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentVersionSearchResult'
        )
        DocumentVersionPageSearchResult = apps.get_model(
            app_label='documents',
            model_name='DocumentVersionPageSearchResult'
        )

        DocumentTag = self.get_model(model_name='DocumentTag')
        Tag = self.get_model(model_name='Tag')

        DynamicSerializerField.add_serializer(
            klass=Tag,
            serializer_class='mayan.apps.tags.serializers.TagSerializer'
        )

        Document.add_to_class(
            name='get_tags', value=method_document_get_tags
        )

        EventModelRegistry.register(model=Tag)

        ModelCopy(
            model=Tag, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'label', 'color', 'documents'
            )
        )

        ModelEventType.register(
            model=Document, event_types=(
                event_tag_attached, event_tag_removed
            )
        )
        ModelEventType.register(
            model=Tag, event_types=(
                event_tag_attached, event_tag_edited, event_tag_removed
            )
        )

        ModelFieldRelated(model=Document, name='tags__label')
        ModelFieldRelated(model=Document, name='tags__color')

        ModelPermission.register(
            model=Document, permissions=(
                permission_tag_attach, permission_tag_remove,
                permission_tag_view
            )
        )

        ModelPermission.register(
            model=Tag, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_tag_attach, permission_tag_delete,
                permission_tag_edit, permission_tag_remove,
                permission_tag_view
            )
        )

        model_query_fields_document = ModelQueryFields.get(model=Document)
        model_query_fields_document.add_prefetch_related_field(
            field_name='tags'
        )

        model_query_fields_tag = ModelQueryFields.get(model=Tag)
        model_query_fields_tag.add_prefetch_related_field(
            field_name='documents'
        )

        # Document

        SourceColumn(
            label=_(message='Tags'), source=Document, widget=DocumentTagWidget
        )

        SourceColumn(
            attribute='document', label=_(message='Tags'),
            source=DocumentFileSearchResult, widget=DocumentTagWidget
        )
        SourceColumn(
            attribute='document_file__document', label=_(message='Tags'),
            source=DocumentFilePageSearchResult, widget=DocumentTagWidget
        )

        SourceColumn(
            attribute='document', label=_(message='Tags'),
            source=DocumentVersionSearchResult, widget=DocumentTagWidget
        )
        SourceColumn(
            attribute='document_version__document', label=_(message='Tags'),
            source=DocumentVersionPageSearchResult, widget=DocumentTagWidget
        )

        # Tag

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=DocumentTag
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=Tag
        )
        SourceColumn(
            attribute='get_preview_widget', include_label=True, source=Tag
        )
        source_column_tag_document_count = SourceColumn(
            func=lambda context: context['object'].get_document_count(
                user=context['request'].user
            ), include_label=True, label=_(message='Documents'), source=Tag
        )
        source_column_tag_document_count.add_exclude(source=DocumentTag)

        # Document

        menu_list_facet.bind_links(
            links=(link_document_tag_list,), sources=(Document,)
        )

        menu_multi_item.bind_links(
            links=(
                link_document_multiple_tag_multiple_attach,
                link_document_multiple_tag_multiple_remove
            ),
            sources=(Document,)
        )

        # Tag

        menu_list_facet.bind_links(
            links=(
                link_tag_document_list,
            ), sources=(Tag,)
        )

        menu_tags.bind_links(
            links=(
                link_tag_list, link_tag_create
            )
        )

        menu_main.bind_links(links=(menu_tags,), position=60)

        menu_multi_item.bind_links(
            exclude=(DocumentTag,),
            links=(link_tag_multiple_delete,), sources=(Tag,)
        )

        menu_object.bind_links(
            exclude=(DocumentTag,),
            links=(
                link_tag_edit, link_tag_single_delete
            ),
            sources=(Tag,)
        )

        menu_secondary.bind_links(
            links=(
                link_document_tag_multiple_attach,
                link_document_tag_multiple_remove
            ),
            sources=(
                'tags:tag_attach', 'tags:document_tag_list',
                'tags:single_document_multiple_tag_remove'
            )
        )

        # Index update

        post_save.connect(
            dispatch_uid='tags_handler_index_document',
            receiver=handler_index_document,
            sender=Tag
        )
        pre_delete.connect(
            dispatch_uid='tags_handler_tag_pre_delete',
            receiver=handler_tag_pre_delete,
            sender=Tag
        )



from datetime import datetime

from mayan.apps.testing.tests.base import BaseTestCase

from .literals import TEST_TEMPLATE_TAG_RESULT
from .mixins import TemplateTestMixin


class TemplateTagDateParseTestCase(TemplateTestMixin, BaseTestCase):
    def test_basic_functionality(self):
        now = datetime.now()

        result = self._render_test_template(
            template_string='{{% set "{}"|date_parse as date_object %}}{{{{ date_object.year }}}}'.format(
                now.isoformat()
            )
        )
        self.assertEqual(
            result, str(now.year)
        )


class TemplateFilterDictGetTestCase(TemplateTestMixin, BaseTestCase):
    def test_filter_dict_get_valid(self):
        result = self._render_test_template(
            template_string='{{ dict|dict_get:1 }}', context={
                'dict': {1: 'a'}
            }
        )
        self.assertEqual(result, 'a')

    def test_filter_dict_get_invalid(self):
        result = self._render_test_template(
            template_string='{{ dict|dict_get:2 }}', context={
                'dict': {1: 'a'}
            }
        )
        self.assertEqual(result, '')


class TemplateFilterSplitTestCase(TemplateTestMixin, BaseTestCase):
    def test_filter_split_valid(self):
        result = self._render_test_template(
            template_string='{% with x|split:"," as result %}{{ result.0 }}-{{ result.1 }}-{{ result.2 }}{% endwith %}', context={'x': '1,2,3'}
        )
        self.assertEqual(result, '1-2-3')


class TemplateTagLoadingTestCase(TemplateTestMixin, BaseTestCase):
    def test_user_template_tag_loading(self):
        result = self._render_test_template(
            template_string='{% load templating_test_tags %}{% templating_test_tag %}'
        )
        self.assertEqual(result, TEST_TEMPLATE_TAG_RESULT)


class TemplateTagRegexTestCase(TemplateTestMixin, BaseTestCase):
    def test_tag_regex_findall_false(self):
        result = self._render_test_template(
            template_string='{% regex_findall "\\d" "abcxyz" as result %}{% if result %}{{ result }}{% endif %}'
        )
        self.assertEqual(result, '')

    def test_tag_regex_findall_true(self):
        result = self._render_test_template(
            template_string='{% regex_findall "\\d" "abc123" as result %}{{ result.0 }}{{ result.1 }}{{ result.2 }}'
        )
        self.assertEqual(result, '123')

    def test_tag_regex_match_false(self):
        result = self._render_test_template(
            template_string='{% regex_match "\\d" "abc123" as result %}{% if result %}{{ result }}{% endif %}'
        )
        self.assertEqual(result, '')

    def test_tag_regex_match_true(self):
        result = self._render_test_template(
            template_string='{% regex_match "\\d" "123abc" as result %}{% if result %}{{ result.0 }}{% endif %}'
        )
        self.assertEqual(result, '1')

    def test_tag_regex_search_false(self):
        result = self._render_test_template(
            template_string='{% regex_search "\\d" "abcxyz" as result %}{% if result %}{{ result }}{% endif %}'
        )
        self.assertEqual(result, '')

    def test_tag_regex_search_true(self):
        result = self._render_test_template(
            template_string='{% regex_search "\\d" "abc123" as result %}{{ result.0 }}'
        )
        self.assertEqual(result, '1')

    def test_tag_regex_sub_false(self):
        result = self._render_test_template(
            template_string='{% regex_sub "\\d" "XX" "abcxyz" as result %}{{ result }}'
        )
        self.assertEqual(result, 'abcxyz')

    def test_tag_regex_sub_true(self):
        result = self._render_test_template(
            template_string='{% regex_sub "\\d" "XX" "abc123" as result %}{{ result }}'
        )
        self.assertEqual(result, 'abcXXXXXX')


class TemplateTagSetTestCase(TemplateTestMixin, BaseTestCase):
    def test_tag_set_string(self):
        result = self._render_test_template(
            template_string='{% set "string" as result %}{{ result }}'
        )
        self.assertEqual(result, 'string')

    def test_tag_set_number(self):
        result = self._render_test_template(
            template_string='{% set 99 as result %}{{ result }}'
        )
        self.assertEqual(result, '99')

    def test_tag_set_logical(self):
        result = self._render_test_template(
            template_string='{% set True as result %}{{ result }}'
        )
        self.assertEqual(result, 'True')

    def test_tag_set_nonexistant(self):
        result = self._render_test_template(
            template_string='{% set nonexistent as result %}{{ result }}'
        )
        self.assertEqual(result, '')


class TemplateTagTimeDeltaTestCase(TemplateTestMixin, BaseTestCase):
    def test_basic_functionality(self):
        now = datetime.now()

        result = self._render_test_template(
            template_string='{{% set "{}"|date_parse as date_object %}}{{% timedelta date_object days=366 as date_new %}}{{{{ date_new.year }}}}'.format(
                now.isoformat()
            )
        )
        self.assertEqual(
            result, str(now.year + 1)
        )



from django import forms
from django.utils.translation import gettext_lazy as _

from ..models import WorkflowTransition


class WorkflowInstanceTransitionSelectForm(forms.Form):
    def __init__(self, user, workflow_instance, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields[
            'transition'
        ].queryset = workflow_instance.get_transition_choices(user=user)

    transition = forms.ModelChoiceField(
        help_text=_(
            message='Select a transition to execute in the next step.'
        ), label=_(message='Transition'),
        queryset=WorkflowTransition.objects.none()
    )



import hashlib
import io
import logging

import cairosvg
from furl import furl
from PIL import Image

from django.shortcuts import reverse
from django.utils.encoding import force_bytes
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from mayan.apps.file_caching.models import Cache, CachePartitionFile
from mayan.apps.templating.classes import Template

from .literals import STORAGE_NAME_SIGNATURE_CAPTURES_CACHE

logger = logging.getLogger(name=__name__)


class SignatureCaptureBusinessLogicMixin:
    @cached_property
    def cache(self):
        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_SIGNATURE_CAPTURES_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='{}'.format(self.pk)
        )
        return partition

    def generate_image(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        # The parameters 'maximum_layer_order',
        # `transformation_instance_list`, `user` are not used, but added
        # to retain interface compatibility.
        cache_filename = '{}'.format(
            self.get_hash()
        )

        try:
            self.cache_partition.get_file(filename=cache_filename)
        except CachePartitionFile.DoesNotExist:
            logger.debug(
                'signature capture cache file "%s" not found', cache_filename
            )

            image = self.get_image()
            with io.BytesIO() as image_buffer:
                image.save(image_buffer, format='PNG')

                with self.cache_partition.create_file(filename=cache_filename) as file_object:
                    file_object.write(
                        image_buffer.getvalue()
                    )
        else:
            logger.debug(
                'signature_capture cache file "%s" found', cache_filename
            )

        return cache_filename

    def get_api_image_url(self, *args, **kwargs):
        final_url = furl()
        final_url.args = kwargs
        final_url.path = reverse(
            viewname='rest_api:signature_capture-image',
            kwargs={
                'document_id': self.document.pk,
                'signature_capture_id': self.pk
            }
        )
        final_url.args['_hash'] = self.get_hash()

        return final_url.tostr()

    def get_date_time_created(self):
        return Template(
            template_string='{{ instance.date_time_created }}'
        ).render(
            context={'instance': self}
        )
    get_date_time_created.short_description = _(message='Creation date and time')

    def get_hash(self):
        return hashlib.sha256(
            string=force_bytes(self.svg)
        ).hexdigest()

    def get_image(self):
        stream = io.BytesIO()
        cairosvg.svg2png(url=self.svg, write_to=stream)
        image = Image.open(stream)

        return image



from django.core.exceptions import ImproperlyConfigured


class DynamicFormBackendMixin:
    """
    The form_fields attribute is a list of dictionaries with the format:
    {
        'name': ''  # Field internal name
        'label': ''  # Label to show to users
        'initial': ''  # Field initial value
        'default': ''  # Default value.
    }
    """
    form_field_widgets = {}
    form_fields = {}
    form_fieldset_exclude_list = ('backend_data',)
    form_media = {}

    @classmethod
    def do_sanity_check(cls, attribute, name):
        super_attribute = getattr(
            cls, name, {}
        )

        if id(super_attribute) == id(attribute):
            raise ImproperlyConfigured(
                'Don\'t update the form `{}` dictionary `{}` directly. '
                'Instead copy the dictionary.'.format(cls, name)
            )

    @classmethod
    def get_form_field_order(cls):
        return getattr(
            cls, 'form_field_order', ()
        )

    @classmethod
    def get_form_field_widgets(cls):
        form_field_widgets = getattr(
            cls, 'form_field_widgets', {}
        )

        form_field_widgets_copy = form_field_widgets.copy()

        return form_field_widgets_copy

    @classmethod
    def get_form_fields(cls):
        form_fields = getattr(
            cls, 'form_fields', {}
        )
        form_fields_copy = form_fields.copy()

        return form_fields_copy

    @classmethod
    def get_form_fieldset_exclude_list(cls):
        return getattr(
            cls, 'form_fieldset_exclude_list', ()
        )

    @classmethod
    def get_form_fieldsets(cls):
        return getattr(
            cls, 'form_fieldsets', ()
        )

    @classmethod
    def get_form_media(cls):
        form_media = getattr(
            cls, 'form_media', {}
        )
        form_media_copy = form_media.copy()

        return form_media_copy

    @classmethod
    def get_form_schema(cls, *args, **kwargs):
        form_fields = cls.get_form_fields(**kwargs)
        form_fieldset_exclude_list = cls.get_form_fieldset_exclude_list(**kwargs)
        form_fieldsets = cls.get_form_fieldsets(**kwargs)
        form_media = cls.get_form_media()
        form_field_widgets = cls.get_form_field_widgets(**kwargs)

        result = {
            'fields': form_fields,
            'fieldset_exclude_list': form_fieldset_exclude_list,
            'fieldsets': form_fieldsets,
            'media': form_media,
            'widgets': form_field_widgets
        }

        form_field_order = cls.get_form_field_order(**kwargs) or tuple(
            form_fields.keys()
        )

        result['field_order'] = form_field_order

        cls.do_sanity_check(attribute=form_fields, name='form_fields')
        cls.do_sanity_check(attribute=form_media, name='form_media')
        cls.do_sanity_check(
            attribute=form_field_widgets, name='form_field_widgets'
        )

        return result



from django.utils.translation import gettext_lazy as _

from ..icons import icon_document_recently_created_list
from ..models.document_models import RecentlyCreatedDocument

from .document_views import DocumentListView


class RecentCreatedDocumentListView(DocumentListView):
    view_icon = icon_document_recently_created_list

    def get_document_queryset(self):
        return RecentlyCreatedDocument.valid.all()

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_icon': icon_document_recently_created_list,
                'no_results_text': _(
                    'This view will list the latest documents created '
                    'in the system.'
                ),
                'no_results_title': _(
                    'There are no recently created documents'
                ),
                'title': _(message='Recently created')
            }
        )
        return context



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_file_mixins import DocumentFileTestMixin
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_document_file_downloaded
from ..permissions import permission_document_file_download

from .mixins import DocumentFileDownloadAPIViewTestMixin


class DocumentFileDownloadAPIViewTestCase(
    DocumentFileDownloadAPIViewTestMixin, DocumentTestMixin,
    DocumentFileTestMixin, BaseAPITestCase
):
    def test_document_file_download_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_download_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_download_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_download
        )

        self._clear_events()

        response = self._request_test_document_file_download_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        with self._test_document.file_latest.open() as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
                filename=self._test_document.file_latest.filename,
                mime_type=self._test_document.file_latest.mimetype
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_document_file_downloaded.id)

    def test_trashed_document_file_download_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_download
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_download_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.acls.classes import ModelPermission
from mayan.apps.converter.models import LayerTransformation
from mayan.apps.converter.tests.mixins import TransformationViewTestMixin
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..layers import layer_redactions
from ..permissions import (
    permission_redaction_create, permission_redaction_delete,
    permission_redaction_edit, permission_redaction_view
)
from ..transformations import TransformationRedactionPercent

from .literals import TEST_REDACTION_ARGUMENT, TEST_REDACTION_ARGUMENT_EDITED


class RedactionViewTestCase(
    TransformationViewTestMixin, GenericDocumentViewTestCase
):
    _test_layer = layer_redactions
    _test_transformation_argument = TEST_REDACTION_ARGUMENT
    _test_transformation_argument_edited = TEST_REDACTION_ARGUMENT_EDITED
    auto_create_test_layer = False
    auto_create_test_transformation_class = False
    TestTransformationClass = TransformationRedactionPercent

    def setUp(self):
        super().setUp()

        ModelPermission.register(
            model=self._test_transformation_object_parent._meta.model, permissions=(
                permission_redaction_create,
                permission_redaction_delete,
                permission_redaction_edit,
                permission_redaction_view
            )
        )

    def test_redaction_create_post_view_no_permission(self):
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_transformation_create_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_create_post_view_with_permission(self):
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_create
        )
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_transformation_create_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_create_get_view_no_permission(self):
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_transformation_create_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_create_get_view_with_permission(self):
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_create
        )
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_transformation_create_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_delete_view_no_permission(self):
        self._create_test_transformation()
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_transformation_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_delete_view_with_access(self):
        self._create_test_transformation()
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_delete
        )
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_transformation_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_edit_view_no_permission(self):
        self._create_test_transformation()
        transformation_arguments = self._test_transformation.arguments

        self._clear_events()

        response = self._request_transformation_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_transformation.refresh_from_db()
        self.assertEqual(
            transformation_arguments, self._test_transformation.arguments
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_edit_view_with_access(self):
        self._create_test_transformation()
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_edit
        )
        transformation_arguments = self._test_transformation.arguments

        self._clear_events()

        response = self._request_transformation_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_transformation.refresh_from_db()
        self.assertNotEqual(
            transformation_arguments, self._test_transformation.arguments
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_list_view_no_permission(self):
        self._create_test_transformation()

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_transformation_object_parent), status_code=404
        )
        self.assertNotContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_list_view_with_access(self):
        self._create_test_transformation()
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_view
        )

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object_parent), status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_select_get_view_no_permission(self):
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_test_transformation_select_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_select_get_view_with_access(self):
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_create
        )
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_test_transformation_select_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_select_post_view_no_permission(self):
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_test_transformation_select_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_redaction_select_post_view_with_access(self):
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=permission_redaction_create
        )
        transformation_count = LayerTransformation.objects.count()

        self._clear_events()

        response = self._request_test_transformation_select_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            LayerTransformation.objects.count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b, worker_c

queue_file_caching = CeleryQueue(
    name='file_caching', label=_(message='File caching'), worker=worker_b
)
queue_file_caching_slow = CeleryQueue(
    name='file_caching_slow', label=_(message='File caching slow'), worker=worker_c
)

queue_file_caching.add_task_type(
    dotted_path='mayan.apps.file_caching.tasks.task_cache_partition_purge',
    label=_(message='Purge a file cache partition')
)

queue_file_caching_slow.add_task_type(
    dotted_path='mayan.apps.file_caching.tasks.task_cache_purge',
    label=_(message='Purge a file cache')
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('common', '0013_auto_20190725_0452')
    ]

    operations = [
        migrations.AlterField(
            model_name='userlocaleprofile',
            name='language',
            field=models.CharField(
                choices=[
                    ('ar', 'Arabic'), ('bg', 'Bulgarian'),
                    ('bs', 'Bosnian'), ('cs', 'Czech'), ('da', 'Danish'),
                    ('de', 'German'), ('el', 'Greek'), ('en', 'English'),
                    ('es', 'Spanish'), ('fa', 'Persian'), ('fr', 'French'),
                    ('hu', 'Hungarian'), ('id', 'Indonesian'),
                    ('it', 'Italian'), ('lv', 'Latvian'),
                    ('nl', 'Dutch'), ('pl', 'Polish'),
                    ('pt', 'Portuguese'), ('pt-br', 'Portuguese (Brazil)'),
                    ('ro', 'Romanian'), ('ru', 'Russian'),
                    ('sl', 'Slovenian'), ('tr', 'Turkish'),
                    ('vi', 'Vietnamese'), ('zh-hans', 'Chinese (Simplified)')
                ], max_length=8, verbose_name='Language'
            ),
        )
    ]



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _, ngettext

from mayan.apps.documents.forms.document_type_forms import DocumentTypeFilteredSelectForm
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.models.document_version_models import DocumentVersion
from mayan.apps.documents.models.document_version_page_models import DocumentVersionPage
from mayan.apps.views.generics import (
    FormView, MultipleObjectConfirmActionView, MultipleObjectDeleteView,
    SingleObjectDetailView, SingleObjectDownloadView, SingleObjectEditView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .forms import (
    DocumentVersionPageOCRContentDetailForm,
    DocumentVersionPageOCRContentEditForm, DocumentVersionOCRContentForm
)
from .icons import (
    icon_document_type_ocr_settings, icon_document_type_ocr_submit,
    icon_document_version_ocr_content_single_delete,
    icon_document_version_ocr_content_detail,
    icon_document_version_ocr_content_download,
    icon_document_version_ocr_single_submit,
    icon_document_version_page_ocr_content_detail,
    icon_document_version_page_ocr_content_edit
)
from .models import DocumentVersionPageOCRContent
from .permissions import (
    permission_document_version_ocr_content_edit,
    permission_document_version_ocr_content_view,
    permission_document_version_ocr, permission_document_type_ocr_setup
)


class DocumentTypeOCRSettingsEditView(
    ExternalObjectViewMixin, SingleObjectEditView
):
    external_object_class = DocumentType
    external_object_permission = permission_document_type_ocr_setup
    external_object_pk_url_kwarg = 'document_type_id'
    fields = ('auto_ocr',)
    post_action_redirect = reverse_lazy(
        viewname='documents:document_type_list'
    )
    view_icon = icon_document_type_ocr_settings

    def get_document_type(self):
        return self.external_object

    def get_extra_context(self):
        return {
            'object': self.get_document_type(),
            'title': _(
                'Edit OCR settings for document type: %s.'
            ) % self.get_document_type()
        }

    def get_object(self, queryset=None):
        return self.get_document_type().ocr_settings


class DocumentTypeOCRSubmitView(FormView):
    extra_context = {
        'title': _(message='Submit all documents of a type for OCR')
    }
    form_class = DocumentTypeFilteredSelectForm
    post_action_redirect = reverse_lazy(viewname='common:tools_list')
    view_icon = icon_document_type_ocr_submit

    def form_valid(self, form):
        count = 0

        queryset_valid_documents = Document.valid.all()

        for document_type in form.cleaned_data['document_type']:
            queryset = document_type.documents.filter(
                pk__in=queryset_valid_documents.values('pk')
            )
            for document in queryset:
                document.submit_for_ocr(user=self.request.user)
                count += 1

        messages.success(
            message=_(
                '%(count)d documents added to the OCR queue.'
            ) % {
                'count': count
            }, request=self.request
        )

        return HttpResponseRedirect(
            redirect_to=self.get_success_url()
        )

    def get_form_extra_kwargs(self):
        return {
            'allow_multiple': True,
            'permission': permission_document_version_ocr,
            'user': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(viewname='common:tools_list')


class DocumentVersionOCRContentDeleteView(MultipleObjectDeleteView):
    error_message = _(
        'Error deleting document version OCR "%(instance)s"; %(exception)s'
    )
    object_permission = permission_document_version_ocr
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid.all()
    success_message_plural = _(
        'OCR content of %(count)d document versions deleted successfully.'
    )
    success_message_single = _(
        'OCR content of "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        'OCR content of %(count)d document version deleted successfully.'
    )
    title_plural = _(
        'Delete the OCR content of the %(count)d selected document versions.'
    )
    title_single = _(message='Delete the OCR content of: %(object)s.')
    title_singular = _(
        'Delete the OCR content of the %(count)d selected document version.'
    )
    view_icon = icon_document_version_ocr_content_single_delete

    def object_action(self, form, instance):
        DocumentVersionPageOCRContent.objects.delete_content_for(
            document_version=instance, user=self.request.user
        )


class DocumentVersionOCRContentView(SingleObjectDetailView):
    form_class = DocumentVersionOCRContentForm
    object_permission = permission_document_version_ocr_content_view
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid.all()
    view_icon = icon_document_version_ocr_content_detail

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(
            request, *args, **kwargs
        )
        self.object.document.add_as_recent_document_for_user(
            user=request.user
        )
        return result

    def get_extra_context(self):
        return {
            'document': self.object,
            'hide_labels': True,
            'object': self.object,
            'title': _(message='OCR result for document: %s') % self.object
        }


class DocumentVersionOCRDownloadView(SingleObjectDownloadView):
    object_permission = permission_document_version_ocr_content_view
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid.all()
    view_icon = icon_document_version_ocr_content_download

    def get_download_file_object(self):
        return self.object.ocr_content()

    def get_download_filename(self):
        return '{}-OCR'.format(self.object)


class DocumentVersionOCRSubmitView(MultipleObjectConfirmActionView):
    object_permission = permission_document_version_ocr
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid.all()
    success_message = _(
        '%(count)d document version submitted to the OCR queue.'
    )
    success_message_plural = _(
        '%(count)d document versions submitted to the OCR queue.'
    )
    view_icon = icon_document_version_ocr_single_submit

    def get_extra_context(self):
        queryset = self.object_list

        result = {
            'title': ngettext(
                singular='Submit the selected document version to the OCR queue?',
                plural='Submit the selected document versions to the OCR queue?',
                number=queryset.count()
            )
        }

        if queryset.count() == 1:
            result['object'] = queryset.first()

        return result

    def object_action(self, form, instance):
        instance.submit_for_ocr(user=self.request.user)


class DocumentVersionPageOCRContentDetailView(SingleObjectDetailView):
    form_class = DocumentVersionPageOCRContentDetailForm
    object_permission = permission_document_version_ocr_content_view
    pk_url_kwarg = 'document_version_page_id'
    source_queryset = DocumentVersionPage.valid.all()
    view_icon = icon_document_version_page_ocr_content_detail

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(
            request, *args, **kwargs
        )
        self.object.document_version.document.add_as_recent_document_for_user(
            user=request.user
        )
        return result

    def get_extra_context(self):
        return {
            'hide_labels': True,
            'object': self.object,
            'title': _(
                'OCR result for document version page: %s'
            ) % self.object
        }


class DocumentVersionPageOCRContentEditView(
    ExternalObjectViewMixin, SingleObjectEditView
):
    external_object_queryset = DocumentVersionPage.valid.all()
    external_object_permission = permission_document_version_ocr_content_edit
    external_object_pk_url_kwarg = 'document_version_page_id'
    form_class = DocumentVersionPageOCRContentEditForm
    view_icon = icon_document_version_page_ocr_content_edit

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(
            request, *args, **kwargs
        )
        self.external_object.document_version.document.add_as_recent_document_for_user(
            user=request.user
        )
        return result

    def get_extra_context(self):
        return {
            'hide_labels': True,
            'object': self.external_object,
            'title': _(
                'Edit OCR for document version page: %s'
            ) % self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_object(self):
        try:
            return self.external_object.ocr_content
        except DocumentVersionPageOCRContent.DoesNotExist:
            return DocumentVersionPageOCRContent(
                document_version_page=self.external_object
            )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0033_auto_20160325_0052')
    ]

    operations = [
        migrations.AlterField(
            model_name='document',
            name='in_trash',
            field=models.BooleanField(
                default=False, verbose_name='In trash?', db_index=True,
                editable=False
            )
        ),
        migrations.AlterField(
            model_name='document',
            name='is_stub',
            field=models.BooleanField(
                default=True, editable=False,
                help_text='A document stub is a document with an entry on '
                'the database but no file uploaded. This could be an '
                'interrupted upload or a deferred upload via the API.',
                verbose_name='Is stub?', db_index=True
            )
        )
    ]



from mayan.apps.rest_api import generics

from ..models.recently_accessed_document_models import RecentlyAccessedDocument
from ..permissions import permission_document_view
from ..serializers.recently_accessed_document_serializers import (
    RecentlyAccessedDocumentSerializer
)


class APIRecentlyAccessedDocumentListView(generics.ListAPIView):
    """
    get: Return a list of the recently accessed documents for the current user.
    """
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = RecentlyAccessedDocumentSerializer

    def get_source_queryset(self):
        return RecentlyAccessedDocument.valid.filter(user=self.request.user)



import unicodedata


def generate_username_from_email(email):
    # Using Python 3 and Django 1.11+, usernames can contain alphanumeric
    # (ascii and unicode), _, @, +, . and - characters. So we normalize
    # it and slice at 150 characters.
    return unicodedata.normalize('NFKC', email)[:150]



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0010_auto_20151001_0055')
    ]

    operations = [
        migrations.CreateModel(
            name='SaneScanner',
            fields=[
                (
                    'interactivesource_ptr', models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True, primary_key=True, serialize=False,
                        to='sources.InteractiveSource'
                    )
                ),
                (
                    'device_name', models.CharField(
                        help_text='Device name as returned by the SANE '
                        'backend.', max_length=255,
                        verbose_name='Device name'
                    )
                )
            ],
            options={
                'verbose_name': 'SANE Scanner',
                'verbose_name_plural': 'SANE Scanners'
            },
            bases=('sources.interactivesource',),
        )
    ]
